import re
from typing import (
    Iterator,
    List,
    Optional,
    Union,
    TypeVar,
    Generic,
    Type,
    Annotated,
    Any,
)
from datetime import datetime
from uuid import UUID
from enum import Enum
from concurrent.futures import ThreadPoolExecutor

from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    WithJsonSchema,
    SerializerFunctionWrapHandler,
    WrapSerializer,
    SerializationInfo,
)

from fastapi import Path
import cattrs.preconf.json

converter = cattrs.preconf.json.make_converter()
converter.register_unstructure_hook(
    cls=int, func=lambda x: str(x) if x > 2**53 else x
)

optional_field = Field(default=None, json_schema_extra=lambda x: x.pop("default"))
optional_deprecated_field = Field(
    default=None, deprecated=True, json_schema_extra=lambda x: x.pop("default")
)


def special_dict_wrap(
    v: Any, nxt: SerializerFunctionWrapHandler, info: SerializationInfo
):
    if info.context and info.context.get("stringify_bigints", False):
        return converter.unstructure(v)
    return nxt(v)


def big_int_wrap(
    v: Any, nxt: SerializerFunctionWrapHandler, info: SerializationInfo
) -> Union[str, int]:
    if info.context and info.context.get("stringify_bigints", False):
        return str(v)
    return nxt(v)


SpecialDict = Annotated[
    dict,
    dict,
    WrapSerializer(special_dict_wrap, when_used="json"),
]


Int64 = Annotated[
    int,
    int,
    WithJsonSchema({"type": "integer", "format": "int64"}),
    WrapSerializer(big_int_wrap, when_used="json"),
    Path(json_schema_extra={"type": "integer", "format": "int64"}),
]


def optional_field_alt(description: str):
    return Field(
        default=None,
        json_schema_extra=lambda x: x.pop("default"),
        description=description,
    )


def convert_camel_to_snake_case(camel_string: str) -> str:
    return re.sub("(.)([A-Z])", r"\1_\2", camel_string).lower()


class EmptyModel(BaseModel):
    pass


T = TypeVar("T", bound="ResourceModel")


class ResourceModel(BaseModel, Generic[T]):
    _app: Any = None
    model_config = ConfigDict(from_attributes=True)

    @classmethod
    def _describe(cls, return_description: bool = False):
        description = f"{cls.__name__}:\n"
        model_field_items = cls.model_fields.items()

        # reorganize the fields so that the common fields are at the top
        top_fields = ("id", "timestamp")
        bottom_fields = (
            "created_at",
            "updated_at",
            "deleted_at",
            "created_by",
            "updated_by",
            "deleted_by",
        )
        model_field_items = sorted(
            model_field_items,
            key=lambda x: (
                0 if x[0] in top_fields else 1,
                1 if x[0] in bottom_fields else 0,
                x[0],
            ),
        )

        for field_name, field_info in model_field_items:
            if field_name in ResourceModel.model_fields:
                continue

            field_type = field_info.annotation
            if hasattr(field_type, "__args__"):
                raw_sub_type_name = str(field_type.__args__[0])
                if "typing.Annotated[int" in raw_sub_type_name:
                    sub_type_name = "int"
                else:
                    sub_type_name = raw_sub_type_name
                if str(field_type).startswith("typing."):
                    type_name = str(field_type)[7:]
                else:
                    type_name = field_type.__name__
                type_name = type_name.replace(raw_sub_type_name, sub_type_name)
            elif str(field_type).startswith("typing."):
                type_name = str(field_type)[7:]
            else:
                type_name = field_type.__name__
            type_name = re.sub(r"(\w+\.)+", "", type_name)

            field_title = f"{field_name} [{type_name}]"
            buffer_length = 38 - len(field_title)
            description += (
                f"{field_title}: {' ' * buffer_length}{field_info.description or ''}\n"
            )
        description = description[:-1]
        if return_description:
            return description
        else:
            print(description)

    _repr_fields = ("id", "name")

    def __repr__(self):
        model_name = self.__class__.__name__
        value = f"<{model_name} "
        for field_name in self._repr_fields:
            if field_name in self.model_fields:
                value += f"{field_name}={getattr(self, field_name)}, "
        value = value[:-2] + ">"
        return value

    @classmethod
    def set_app(cls, app):
        cls._app = app

    @classmethod
    def fetch(cls: Type[T], *args, **kwargs) -> T:
        id = None
        friendly_id = None
        parent_id = None
        if len(args) > 0:
            try:
                if type(args[0]) is UUID:
                    id = args[0]
                elif type(args[0]) is int:
                    id = int(args[0])
                else:
                    id = UUID(args[0])
            except ValueError:
                friendly_id = args[0]
            if len(args) > 1:
                try:
                    if type(args[1]) is UUID:
                        parent_id = args[1]
                    else:
                        parent_id = UUID(args[1])
                except ValueError:
                    raise ValueError(
                        "Second argument must be the UUID of the parent resource."
                    )

        if "id" in kwargs:
            id = kwargs["id"]

        if id is not None:
            fetch_method = getattr(
                cls._app.fetch, convert_camel_to_snake_case(cls.__name__)
            )
            if parent_id is not None:
                return fetch_method(id, parent_id, **kwargs).data
            else:
                return fetch_method(id, **kwargs).data
        else:
            list_method = getattr(
                cls._app.list, convert_camel_to_snake_case(cls.__name__)
            )

            list_args = {
                "order": "created_at",
                "sort": "desc",
                "limit": 1,
            }
            list_args.update(kwargs)
            if friendly_id is not None:
                fields = cls._repr_fields.default
                if len(fields) < 2:
                    raise Exception(f"Cannot fetch {cls.__name__} without it's ID.")
                list_args[fields[1]] = friendly_id

            if parent_id is not None:
                list_res = list_method(parent_id, **list_args)
            else:
                list_res = list_method(**list_args)
            if list_res.count == 0:
                raise ValueError(f"No {cls.__name__} found.")
            elif list_res.count > 1:
                try:
                    cls._app.logger.warning(
                        f"Multiple {cls.__name__}s ({list_res.count}) found."
                    )
                except AttributeError:
                    pass
            return list_res.data[0]

    @classmethod
    def first(cls: Type[T], **kwargs) -> T:
        kwargs["order"] = kwargs.get("order", "created_at")
        kwargs["sort"] = kwargs.get("sort", "asc")
        kwargs["limit"] = 1
        return cls.list(**kwargs)[0]

    @classmethod
    def list(cls: Type[T], **kwargs) -> List[T]:
        list_method = getattr(cls._app.list, convert_camel_to_snake_case(cls.__name__))
        return list_method(**kwargs).data

    @classmethod
    def list_all(cls: Type[T], **kwargs) -> List[T]:
        list_method = getattr(cls._app.list, convert_camel_to_snake_case(cls.__name__))
        offset = kwargs.pop("offset", 0)
        limit = kwargs.pop("limit", 100)
        total_count = float("inf")
        resources = []

        while len(resources) < total_count:
            res = list_method(**kwargs, offset=offset, limit=limit)
            total_count = res.count
            resources.extend(res.data)
            offset += limit

        return resources

    @classmethod
    def iterate_efficiently(cls: Type[T], **kwargs) -> Iterator[T]:
        MAX_LIMIT = 100
        order = kwargs.get("order", "created_at")
        assert order == "created_at"
        sort = kwargs.pop("sort", "desc")
        if kwargs.pop("offset", None):
            raise Exception("Can't use offsets with this method")
        limit = kwargs.pop("limit", float("inf"))
        yields_left = limit
        created_at_target = None
        kwargs["include_count"] = False
        while yields_left > 0:
            _limit = min(MAX_LIMIT, yields_left)
            if created_at_target is not None:
                if sort == "desc":
                    created_at_target -= datetime.resolution
                    kwargs["created_at_lte"] = created_at_target
                else:
                    created_at_target += datetime.resolution
                    kwargs["created_at_gte"] = created_at_target
            res = cls.list(**kwargs, limit=_limit, sort=sort)
            if len(res) == 0:
                return
            # this may break when too many records have the same created_at, although it's pretty unlikely
            created_at_target: datetime = res[-1].created_at
            yields_left -= len(res)
            for i in res:
                yield i

    @classmethod
    def iterate(cls: Type[T], **kwargs) -> Iterator[T]:
        MAX_LIMIT = 100
        limit = kwargs.pop("limit", float("inf"))
        if not limit or limit > MAX_LIMIT:
            _limit = MAX_LIMIT
            _offset = kwargs.pop("offset", 0)
            yielded_count = 0
            while yielded_count < limit:
                res = cls.list(**kwargs, limit=_limit, offset=_offset)
                for i in res:
                    yield i
                if len(res) < _limit:
                    return
                _offset += _limit
                yielded_count += limit
        else:
            return iter(cls.list(**kwargs))

    @classmethod
    def create(cls: Type[T], **kwargs) -> T:
        create_method = getattr(
            cls._app.create, convert_camel_to_snake_case(cls.__name__)
        )
        return create_method(**kwargs).data

    def update(self, **kwargs) -> T:
        # get the class name and convert it to snake case
        class_name = convert_camel_to_snake_case(self.__class__.__name__)
        update_method = getattr(self._app.update, class_name)
        updated_resource = update_method(self.id, data=kwargs).data
        for field, value in updated_resource.dict().items():
            # check that the field is a pydantic field
            if field in self.model_fields:
                # set the value of the field
                setattr(self, field, value)
        return self

    @classmethod
    def delete(cls, **kwargs) -> None:
        delete_method = getattr(
            cls._app.delete, convert_camel_to_snake_case(cls.__name__)
        )
        return delete_method(**kwargs)

    @classmethod
    def fetch_or_create(cls, **kwargs):
        try:
            return cls.fetch(**kwargs)
        except Exception:
            return cls.create(**kwargs)

    def refresh(self):
        refreshed_resource = self.fetch(self.id)
        for field, value in refreshed_resource.model_dump().items():
            if field in self.model_fields:
                setattr(self, field, value)
        return self

    def _list_all_subresources(self, list_method, threaded=True, **kwargs):
        if threaded:
            return self._list_all_subresources_threaded(list_method, **kwargs)
        else:
            return self._list_all_subresources_single(list_method, **kwargs)

    def _list_all_subresources_single(self, list_method, **kwargs):
        subresources = []
        total_limit = kwargs.pop("limit", None)
        limit = 100 if total_limit is None else min(100, total_limit)
        offset = 0
        kwargs["limit"] = limit
        kwargs["offset"] = offset
        while True:
            res = list_method(**kwargs)
            subresources.extend(res.data)
            if len(res.data) < limit:
                break
            if total_limit is not None:
                total_limit -= limit
                if total_limit <= 0:
                    break
            offset += limit
            kwargs["offset"] = offset
        return subresources

    def _list_all_subresources_threaded(self, list_method, **kwargs):
        count_kwargs = kwargs.copy()
        count_kwargs["include_count"] = True
        count_kwargs["limit"] = 0
        count_kwargs["offset"] = 0
        count = list_method(**count_kwargs).count
        if count == 0:
            return []

        subresources = []
        total_limit = kwargs.pop("limit", None)
        limit = 100 if total_limit is None else min(100, total_limit)
        offset = 0
        kwargs["limit"] = limit
        kwargs["offset"] = offset
        with ThreadPoolExecutor() as executor:
            futures = []
            while offset < count:
                futures.append(executor.submit(list_method, **kwargs))
                offset += limit
                kwargs["offset"] = offset
                if total_limit is not None:
                    total_limit -= limit
                    if total_limit <= 0:
                        break
            for future in futures:
                res = future.result()
                subresources.extend(res.data)
        return subresources


class CommonModel(ResourceModel[T]):
    id: UUID = Field(..., description="The ID of the resource.")

    created_at: datetime = Field(
        ..., description="The creation timestamp of the resource."
    )
    updated_at: Optional[datetime] = Field(
        ..., description="The last update timestamp of the resource."
    )
    deleted_at: Optional[datetime] = Field(
        ..., description="The deletion timestamp of the resource."
    )
    created_by: Optional[UUID] = Field(
        ..., description="The ID of the user who created the resource."
    )
    updated_by: Optional[UUID] = Field(
        ..., description="The ID of the user who last updated the resource."
    )
    deleted_by: Optional[UUID] = Field(
        ..., description="The ID of the user who deleted the resource."
    )


class TimeSeriesModel(ResourceModel[T]):
    _repr_fields = ("timestamp",)

    timestamp: Int64 = Field(..., description="The timestamp of the resource.")

    created_at: datetime = Field(
        ..., description="The creation timestamp of the resource."
    )
    updated_at: Optional[datetime] = Field(
        ..., description="The last update timestamp of the resource."
    )
    deleted_at: Optional[datetime] = Field(
        ..., description="The deletion timestamp of the resource."
    )
    created_by: Optional[UUID] = Field(
        ..., description="The ID of the user who created the resource."
    )
    updated_by: Optional[UUID] = Field(
        ..., description="The ID of the user who last updated the resource."
    )
    deleted_by: Optional[UUID] = Field(
        ..., description="The ID of the user who deleted the resource."
    )


BaseElement = TypeVar("BaseElement", bound=BaseModel)


class DataResponseModel(BaseModel, Generic[BaseElement]):
    data: BaseElement


class PaginationModel(BaseModel, Generic[BaseElement]):
    offset: int
    limit: int
    order: str
    sort: str
    count: Optional[int]
    data: List[BaseElement]


class PatchOperation(BaseModel):
    op: str
    path: str
    value: Optional[Union[str, int, float, bool, dict, list, None]]


class JSONFilter(BaseModel):
    var: str
    op: str
    val: Union[str, int, float, bool, list, None]


class UploadState(str, Enum):
    complete = "complete"
    processing = "processing"
    aborted = "aborted"



from typing import List, Optional
from datetime import datetime

from pydantic import BaseModel, Field, ConfigDict

from lqs.interface.base.models.__common__ import (
    ResourceModel,
    UploadState,
)


class Object(ResourceModel["Object"]):
    _repr_fields = ("key",)

    key: str = Field(
        ...,
        description="The key of the object (i.e., the path within the object store).",
    )
    etag: Optional[str] = Field(
        None, description="The ETag of the object, used for integrity checks."
    )
    size: Optional[int] = Field(None, description="The size of the object in bytes.")
    last_modified: Optional[datetime] = Field(
        None, description="The last modified timestamp of the object."
    )
    presigned_url: Optional[str] = Field(
        None, description="The presigned URL for fetching the object data."
    )
    upload_state: UploadState = Field(
        ..., description="The upload state of the object."
    )

    model_config = ConfigDict(
        json_schema_extra=lambda x: x["required"].extend(
            [
                "etag",
                "size",
                "last_modified",
                "presigned_url",
            ]
        )
    )


class ObjectDataResponse(BaseModel):
    data: Object


class ObjectListResponse(BaseModel):
    data: List[Object]
    is_truncated: Optional[bool]
    key_count: Optional[int]
    max_keys: int

    continuation_token: Optional[str]
    next_continuation_token: Optional[str]
    prefix: Optional[str]
    start_after: Optional[str]
    delimiter: Optional[str]
    common_prefixes: Optional[List[str]]


class ObjectCreateRequest(BaseModel):
    key: str
    content_type: Optional[str] = None


class ObjectUpdateRequest(BaseModel):
    upload_state: UploadState


# Object Parts


class ObjectPart(BaseModel):
    part_number: int
    etag: str
    size: int
    last_modified: Optional[datetime]
    presigned_url: Optional[str]


class ObjectPartDataResponse(BaseModel):
    data: ObjectPart


class ObjectPartListResponse(BaseModel):
    data: List[ObjectPart]
    part_number_marker: Optional[int]
    next_part_number_marker: Optional[int]
    max_parts: Optional[int]
    is_truncated: Optional[bool]


class ObjectPartCreateRequest(BaseModel):
    part_number: Optional[int] = None
    size: int



from typing import Optional
from uuid import UUID
from datetime import datetime

from pydantic import BaseModel

from lqs.interface.dsm.models.__common__ import (
    CommonModel,
    DataResponseModel,
    PaginationModel,
    optional_field,
)


class Announcement(CommonModel):
    datastore_id: Optional[UUID]
    subject: Optional[str]
    content: Optional[str]
    context: Optional[dict]
    status: Optional[str]
    starts_at: Optional[datetime]
    ends_at: Optional[datetime]


class AnnouncementDataResponse(DataResponseModel[Announcement]):
    pass


class AnnouncementListResponse(PaginationModel[Announcement]):
    pass


class AnnouncementCreateRequest(BaseModel):
    datastore_id: Optional[UUID] = None
    subject: Optional[str] = None
    content: Optional[str] = None
    context: Optional[dict] = None
    status: Optional[str] = None
    starts_at: Optional[datetime] = None
    ends_at: Optional[datetime] = None


class AnnouncementUpdateRequest(BaseModel):
    datastore_id: Optional[UUID] = optional_field
    subject: Optional[str] = optional_field
    content: Optional[str] = optional_field
    context: Optional[dict] = optional_field
    status: Optional[str] = optional_field
    starts_at: Optional[datetime] = optional_field
    ends_at: Optional[datetime] = optional_field



from typing import List, Optional, Tuple
from uuid import UUID

from pydantic import BaseModel, Field, ConfigDict

from lqs.interface.core.models.__common__ import (
    DataResponseModel,
    PaginationModel,
    ProcessModel,
    ProcessCreateRequest,
    ProcessUpdateRequest,
    optional_field,
    Int64,
    process_required_fields,
)
from lqs.interface.core import models


class DigestionPartIndexEntry(BaseModel):
    topic_id: str
    ingestion_id: Optional[str]
    source: Optional[str]

    data_offset: int
    data_length: int

    chunk_compression: Optional[str]
    chunk_offset: Optional[int]
    chunk_length: Optional[int]
    timestamp: Int64


# TODO: in Python 3.11, we can use Tuple[*get_type_hints(DigestionPartIndex).values()]
DigestionPartIndexTuple = Tuple[
    str,
    Optional[str],
    Optional[str],
    int,
    int,
    Optional[str],
    Optional[int],
    Optional[int],
    Int64,
]


class DigestionPartIndex(BaseModel):
    digestion_part_id: UUID
    index: Optional[List[DigestionPartIndexTuple]]


class DigestionPartIndexCreateRequest(BaseModel):
    digestion_part_id: UUID
    index: Optional[List[DigestionPartIndexTuple]] = None


class DigestionPartIndexUpdateRequest(BaseModel):
    index: Optional[List[DigestionPartIndexTuple]] = optional_field


class DigestionPartIndexDataResponse(DataResponseModel[DigestionPartIndex]):
    pass


class DigestionPartIndexListResponse(PaginationModel[DigestionPartIndex]):
    pass


# Digestion Part


class DigestionPart(ProcessModel["DigestionPart"]):
    sequence: int = Field(
        ..., description="The index of the digestion part within the digestion."
    )
    digestion_id: UUID = Field(
        ..., description="The ID of the digestion to which the digestion part belongs."
    )
    context: Optional[dict] = Field(
        None, description="Arbitrary JSON context for the digestion part."
    )
    index: Optional[List[DigestionPartIndexTuple]] = Field(
        None, description="The record metadata index of the digestion part."
    )

    log_id: Optional[UUID] = Field(
        None, description="The ID of the log to which the digestion part belongs."
    )
    group_id: Optional[UUID] = Field(
        None, description="The ID of the group to which the digestion part belongs."
    )

    model_config = ConfigDict(
        json_schema_extra=lambda x: x["required"].extend(
            ["context", "index", "log_id", "group_id"] + process_required_fields
        )
    )

    def fetch_digestion(self) -> "models.Digestion":
        return self._app.fetch.digestion(self.digestion_id).data

    def fetch_log(self) -> "models.Log":
        return self._app.fetch.log(self.log_id).data

    def fetch_group(self) -> "models.Group":
        return self._app.fetch.group(self.group_id).data


class DigestionPartForDigestionCreateRequest(ProcessCreateRequest):
    sequence: int
    context: Optional[dict] = None
    index: Optional[List[DigestionPartIndexTuple]] = None


class DigestionPartCreateRequest(DigestionPartForDigestionCreateRequest):
    digestion_id: UUID


class DigestionPartUpdateRequest(ProcessUpdateRequest):
    sequence: int = optional_field
    context: Optional[dict] = optional_field
    index: Optional[List[DigestionPartIndexTuple]] = optional_field


class DigestionPartDataResponse(DataResponseModel[DigestionPart]):
    pass


class DigestionPartListResponse(PaginationModel[DigestionPart]):
    pass



import struct


SIMPLE_TYPES_DICT = {  # see python module struct
    "int8": "b",
    "uint8": "B",
    # Python 2.6 adds in '?' for C99 _Bool, which appears equivalent to an uint8,
    # thus, we use uint8
    "bool": "B",
    "int16": "h",
    "uint16": "H",
    "int32": "i",
    "uint32": "I",
    "int64": "q",
    "uint64": "Q",
    "float32": "f",
    "float64": "d",
    # deprecated
    "char": "B",  # unsigned
    "byte": "b",  # signed
}


def read_uint8(stream, TRIM_SIZE):
    return struct.unpack("<B", stream.read(1))[0]


def read_uint16(stream, TRIM_SIZE):
    return struct.unpack("<H", stream.read(2))[0]


def read_uint32(stream, TRIM_SIZE):
    return struct.unpack("<I", stream.read(4))[0]


def read_uint64(stream, TRIM_SIZE):
    return struct.unpack("<Q", stream.read(8))[0]


def read_int8(stream, TRIM_SIZE):
    return struct.unpack("<b", stream.read(1))[0]


def read_int16(stream, TRIM_SIZE):
    return struct.unpack("<h", stream.read(2))[0]


def read_int32(stream, TRIM_SIZE):
    return struct.unpack("<i", stream.read(4))[0]


def read_int64(stream, TRIM_SIZE):
    return struct.unpack("<q", stream.read(8))[0]


def read_float32(stream, TRIM_SIZE):
    return struct.unpack("<f", stream.read(4))[0]


def read_float64(stream, TRIM_SIZE):
    return struct.unpack("<d", stream.read(8))[0]


def read_byte(stream, TRIM_SIZE):
    return read_int8(stream, TRIM_SIZE=TRIM_SIZE)


def read_string(stream, TRIM_SIZE):
    length = read_uint32(stream, TRIM_SIZE=TRIM_SIZE)
    if TRIM_SIZE is not None and length > TRIM_SIZE:
        stream.read(length)
        return None
    return stream.read(length).decode("utf-8")


def read_time(stream, TRIM_SIZE):
    return {
        "secs": read_uint32(stream, TRIM_SIZE=TRIM_SIZE),
        "nsecs": read_uint32(stream, TRIM_SIZE=TRIM_SIZE),
    }


def read_duration(stream, TRIM_SIZE):
    return read_time(stream, TRIM_SIZE=TRIM_SIZE)


def read_bool(stream, TRIM_SIZE):
    return read_uint8(stream, TRIM_SIZE=TRIM_SIZE) != 0


def read_char(stream, TRIM_SIZE):
    return read_uint8(stream, TRIM_SIZE=TRIM_SIZE)


def read_array(type, stream, read_element, length, package, TRIM_SIZE):
    if length is None:
        length = read_uint32(stream, TRIM_SIZE=TRIM_SIZE)

    if type in SIMPLE_TYPES_DICT:
        element_size = struct.calcsize(SIMPLE_TYPES_DICT[type])
        array_total_size = element_size * length
        array_bytes = stream.read(array_total_size)
        if TRIM_SIZE is not None and length > TRIM_SIZE:
            return None
        return list(
            struct.unpack(
                f"<{length}{SIMPLE_TYPES_DICT[type]}",
                array_bytes,
            )
        )
    if TRIM_SIZE is not None and length > TRIM_SIZE:
        for i in range(length, TRIM_SIZE):
            read_element(stream, TRIM_SIZE=TRIM_SIZE)
        return None
    return [read_element(stream, TRIM_SIZE=TRIM_SIZE) for i in range(length)]


def read_object(type, stream, current_package, TRIM_SIZE):
    # try to import the type and if that fails try to modify the package and try again
    if "." in type:
        package, class_name = type.rsplit(".")
    else:
        class_name = type
        package = current_package
    module = __import__(package)

    cls = getattr(module, class_name)()
    cls.deserialize(stream, TRIM_SIZE=TRIM_SIZE)
    return cls



from typing import Optional

from pydantic import BaseModel

from lqs.interface.dsm.models.__common__ import (
    CommonModel,
    DataResponseModel,
    PaginationModel,
    optional_field,
)


class Configuration(CommonModel):
    value: dict
    name: Optional[str]
    note: Optional[str]
    default: bool
    disabled: bool


class ConfigurationDataResponse(DataResponseModel[Configuration]):
    pass


class ConfigurationListResponse(PaginationModel[Configuration]):
    pass


class ConfigurationCreateRequest(BaseModel):
    value: dict
    name: Optional[str] = None
    note: Optional[str] = None
    default: bool = False
    disabled: bool = False


class ConfigurationUpdateRequest(BaseModel):
    value: dict = optional_field
    name: Optional[str] = optional_field
    note: Optional[str] = optional_field
    default: bool = optional_field
    disabled: bool = optional_field



from typing import TYPE_CHECKING

from lqs.common.core_utils import CoreUtils

if TYPE_CHECKING:
    from lqs.client import RESTClient


class Utils(CoreUtils):
    def __init__(self, app: "RESTClient"):
        super().__init__(app)



from typing import Optional, TYPE_CHECKING
import requests

import time
import json
import base64
import pprint
import decimal
from datetime import datetime, date
from uuid import UUID
import urllib.parse

if TYPE_CHECKING:
    from lqs.client import RESTClient

from lqs.common import exceptions


class LogQSEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, UUID):
            return obj.hex

        if isinstance(obj, (datetime, date)):
            return obj.isoformat()

        if isinstance(obj, bytes):
            return obj.decode("utf-8")

        if isinstance(obj, decimal.Decimal):
            return float(obj)

        return json.JSONEncoder.default(self, obj)


def output_decorator(func):
    def wrapper(*args, **kwargs):
        if args[0]._pretty:
            return pprint.pprint(func(*args, **kwargs))
        return func(*args, **kwargs)

    return wrapper


class RESTInterface:
    service: str

    def __init__(self, app: "RESTClient"):
        self.app = app
        if self.app.http_client is None:
            self._client = requests
            self._use_content = False
        else:
            self._client = self.app.http_client
            self._use_content = True
        self.version_mismatch_warned = False

    def _get_headers(self, url=None):
        headers = {
            "Content-Type": "application/json",
        }

        api_key_id = self.app.config.api_key_id
        api_key_secret = self.app.config.api_key_secret
        if url is not None and "/apps/dsm/api" in url:
            if (
                self.app.config.dsm_api_key_id is not None
                and self.app.config.dsm_api_key_secret is not None
            ):
                self.app.logger.debug("Using DSM API Key")
                api_key_id = self.app.config.dsm_api_key_id
                api_key_secret = self.app.config.dsm_api_key_secret

        if api_key_id is not None and api_key_secret is not None:
            headers["Authorization"] = "Bearer " + base64.b64encode(
                bytes(
                    f"{api_key_id}:{api_key_secret}",
                    "utf-8",
                )
            ).decode("utf-8")

        headers.update(self.app.config.additional_headers)
        return headers

    def _get_url_param_string(self, args, exclude=[]):
        url_params = ""
        for key, value in args.items():
            if value is not None and key not in ["self"] + exclude:
                if type(value) is dict:
                    value = json.dumps(value, cls=LogQSEncoder)
                elif type(value) is list:
                    if all(isinstance(i, (str, int, float, UUID)) for i in value):
                        for i in value:
                            url_params += f"&{key}={i}"
                        continue
                quoted_value = urllib.parse.quote(str(value))
                url_params += f"&{key}={quoted_value}"
        if len(url_params) > 0:
            url_params = "?" + url_params[1:]
        return url_params

    def _get_payload_data(self, args, exclude=[]):
        payload = {}
        for key, value in args.items():
            if value is not None and key not in ["self"] + exclude:
                payload[key] = value
        return payload

    def _check_version_match(
        self, api_version: Optional[str], client_version: Optional[str]
    ) -> bool:
        if self.app.config.version_match_level is None:
            return True
        if api_version is None:
            return True
        if client_version is None:
            return True

        api_version_parts = api_version.split(".")
        client_version_parts = client_version.split(".")

        if self.app.config.version_match_level.lower() == "major":
            return api_version_parts[0] == client_version_parts[0]
        elif self.app.config.version_match_level.lower() == "minor":
            return (
                api_version_parts[0] == client_version_parts[0]
                and api_version_parts[1] <= client_version_parts[1]
            )

    def _handle_response_data(self, response: requests.Response):
        api_version = response.headers.get("x-logqs-version")
        if api_version is not None:
            versions_match = self._check_version_match(
                api_version=api_version,
                client_version=self.app.version,
            )
            if not versions_match:
                version_mismatch_message = f"Version mismatch: client version: {self.app.version}, server version: {api_version}."
                if self.app.config.raise_for_version_mismatch:
                    raise Exception(version_mismatch_message)
                if self.app.config.warn_for_version_mismatch:
                    if not self.version_mismatch_warned:
                        self.version_mismatch_warned = True
                        self.app.logger.warning(version_mismatch_message)

        if response.status_code == 204:
            return

        content_type = response.headers.get("content-type")
        if content_type == "application/json":
            try:
                response_data = response.json()
            except json.decoder.JSONDecodeError:
                raise Exception(f"Error: {response.text}")
        elif content_type == "text/plain":
            response_data = response.text
        else:
            response_data = response.content

        try:
            response_ok = response.ok
        except AttributeError:
            # the TestClient from starlette does not have an ok attribute
            response_ok = response.is_success

        if response_ok:
            return response_data
        else:
            response_message = response_data
            if isinstance(response_data, bytes):
                raise Exception(f"Error: {response_data}")

            if "message" in response_data:
                response_message = response_data["message"]
            if response.status_code == 400:
                raise exceptions.BadRequestException(msg=response_message)
            if response.status_code == 401:
                raise exceptions.UnauthorizedException(msg=response_message)
            elif response.status_code == 403:
                raise exceptions.ForbiddenException(msg=response_message)
            elif response.status_code == 404:
                raise exceptions.NotFoundException(msg=response_message)
            elif response.status_code == 408:
                raise exceptions.TimeoutException(msg=response_message)
            elif response.status_code == 409:
                raise exceptions.ConflictException(msg=response_message)
            elif response.status_code == 413:
                raise exceptions.ContentTooLargeException(msg=response_message)
            elif response.status_code == 423:
                raise exceptions.LockedException(msg=response_message)
            elif response.status_code == 500:
                raise exceptions.InternalServerErrorException(msg=response_message)
            elif response.status_code == 501:
                raise exceptions.NotImplementedException(msg=response_message)
            raise Exception(response_message)

    def _handle_retries(self, func, retry_count=None):
        if retry_count is None:
            retry_count = self.app.config.retry_count
        for i in range(retry_count + 1):
            try:
                return func()
            except Exception as e:
                if not self.app.config.retry_aggressive:
                    # check if the error is one of our expected exceptions,
                    # i.e., an error that probably won't be resolved by retrying
                    lqs_expected_exceptions = [
                        exceptions.UnauthorizedException,
                        exceptions.ForbiddenException,
                        exceptions.NotFoundException,
                        exceptions.ConflictException,
                        exceptions.ContentTooLargeException,
                        exceptions.LockedException,
                        exceptions.BadRequestException,
                    ]
                    for exception in lqs_expected_exceptions:
                        if isinstance(e, exception):
                            raise e
                if retry_count > 0 and i < retry_count:
                    # exponential backoff
                    backoff = self.app.config.retry_delay * (2**i)
                    self.app.logger.error(f"Error: {e}")
                    self.app.logger.debug(f"Retrying in {backoff} seconds")
                    time.sleep(backoff)
                else:
                    raise e
        raise Exception("Error: Max retries exceeded")

    def _get_url(self, resource_path):
        path_prefix = f"/{self.service}/api"
        datastore_id = self.app.get_datastore_id()
        if self.service == "lqs" and datastore_id is not None:
            path_prefix = f"/{self.service}/{datastore_id}/api"

        url = self.app.config.api_url or ""
        url += self.app.config.api_endpoint_prefix or ""
        url += path_prefix + "/"
        url += resource_path
        return url

    def _head_resource(self, resource_path):
        url = self._get_url(resource_path)
        if self.app.config.dry_run:
            self.app.logger.info(
                {
                    "log_type": "dry_run",
                    "method": "HEAD",
                    "url": url,
                }
            )
            return {}

        def make_request():
            r = self._client.head(
                url,
                headers=self._get_headers(url=url),
                timeout=self.app.config.api_request_timeout,
            )
            response_data = self._handle_response_data(r)
            if isinstance(response_data, (dict, list)):
                return response_data
            else:
                raise Exception("Error: HEAD request returned non-JSON data")

        return self._handle_retries(make_request)

    def _get_resource(
        self,
        resource_path,
        expected_content_type: Optional[str] = "application/json",
        additional_headers=None,
        response_model=None,
    ):
        url = self._get_url(resource_path)
        if self.app.config.dry_run:
            self.app.logger.info(
                {
                    "log_type": "dry_run",
                    "method": "GET",
                    "url": url,
                    "expected_content_type": expected_content_type,
                    "additional_headers": additional_headers,
                }
            )
            return {}

        def make_request():
            headers = self._get_headers(url=url)
            if additional_headers:
                headers = {**headers, **additional_headers}
            r = self._client.get(
                url,
                headers=headers,
                timeout=self.app.config.api_request_timeout,
            )
            response_data = self._handle_response_data(r)
            if expected_content_type == "application/json":
                if isinstance(response_data, (dict, list)):
                    if response_model:
                        return response_model(**response_data)
                    return response_data
                else:
                    raise Exception("Error: GET request returned non-JSON data")
            elif expected_content_type == "text/plain":
                if isinstance(response_data, str):
                    return response_data
                else:
                    raise Exception("Error: GET request returned non-string data")
            else:
                if isinstance(response_data, bytes):
                    return response_data
                else:
                    raise Exception(
                        f"Error: GET request returned non-bytes data: {response_data}"
                    )

        return self._handle_retries(make_request)

    def _create_resource(
        self, resource_path, data, response_model, additiona_params={}
    ):
        url = self._get_url(resource_path)
        if self.app.config.dry_run:
            self.app.logger.info(
                {
                    "log_type": "dry_run",
                    "method": "POST",
                    "url": url,
                    "data": data,
                    "response_model": response_model,
                    "additiona_params": additiona_params,
                }
            )
            return {}

        def make_request():
            if self._use_content:
                params = dict(
                    url=url,
                    params=additiona_params,
                    content=json.dumps(data, cls=LogQSEncoder),
                    headers=self._get_headers(url=url),
                    timeout=self.app.config.api_request_timeout,
                )
            else:
                params = dict(
                    url=url,
                    params=additiona_params,
                    data=json.dumps(data, cls=LogQSEncoder),
                    headers=self._get_headers(url=url),
                    timeout=self.app.config.api_request_timeout,
                )

            r = self._client.post(**params)  # type: ignore
            response_data = self._handle_response_data(r)
            if isinstance(response_data, (dict, list)):
                if response_model is None:
                    return response_data
                return response_model(**response_data)
            else:
                raise Exception("Error: POST request returned non-JSON data")

        return self._handle_retries(make_request)

    def _update_resource(
        self, resource_path, data, response_model, additiona_params={}
    ):
        url = self._get_url(resource_path)
        if self.app.config.dry_run:
            self.app.logger.info(
                {
                    "log_type": "dry_run",
                    "method": "PATCH",
                    "url": url,
                    "data": data,
                    "response_model": response_model,
                    "additiona_params": additiona_params,
                }
            )
            return {}

        def make_request():
            if self._use_content:
                params = dict(
                    url=url,
                    params=additiona_params,
                    content=json.dumps(data, cls=LogQSEncoder),
                    headers=self._get_headers(url=url),
                    timeout=self.app.config.api_request_timeout,
                )
            else:
                params = dict(
                    url=url,
                    params=additiona_params,
                    data=json.dumps(data, cls=LogQSEncoder),
                    headers=self._get_headers(url=url),
                    timeout=self.app.config.api_request_timeout,
                )
            r = self._client.patch(**params)  # type: ignore
            response_data = self._handle_response_data(r)
            if isinstance(response_data, (dict, list)):
                return response_model(**response_data)
            else:
                raise Exception("Error: PATCH request returned non-JSON data")

        return self._handle_retries(make_request)

    def _delete_resource(self, resource_path, additiona_params={}):
        url = self._get_url(resource_path)
        if self.app.config.dry_run:
            self.app.logger.info(
                {
                    "log_type": "dry_run",
                    "method": "DELETE",
                    "url": url,
                    "additiona_params": additiona_params,
                }
            )
            return

        def make_request():
            r = self._client.delete(
                url,
                params=additiona_params,
                headers=self._get_headers(url=url),
                timeout=self.app.config.api_request_timeout,
            )
            response_data = self._handle_response_data(r)
            if response_data is not None:
                raise Exception(f"Error: DELETE request returned data: {response_data}")

        return self._handle_retries(make_request)



from typing import Tuple, Type, Optional

from pydantic_settings import (
    BaseSettings,
    SettingsConfigDict,
    PydanticBaseSettingsSource,
)


class TranscodeConfig(BaseSettings):
    trim_cutoff: Optional[int] = None
    log_level: str = "INFO"
    log_as_json: bool = False

    model_config = SettingsConfigDict(
        env_file=(".env.default", ".env", ".env.local", ".env.dev", ".env.prod"),
        env_prefix="LQS_",
        env_file_encoding="utf-8",
        case_sensitive=False,
        extra="ignore",
    )

    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: Type[BaseSettings],
        init_settings: PydanticBaseSettingsSource,
        env_settings: PydanticBaseSettingsSource,
        dotenv_settings: PydanticBaseSettingsSource,
        file_secret_settings: PydanticBaseSettingsSource,
    ) -> Tuple[PydanticBaseSettingsSource, ...]:
        return (
            init_settings,
            env_settings,
            dotenv_settings,
            file_secret_settings,
        )



from typing import Optional, List
from uuid import UUID

from pydantic import Field, ConfigDict

from lqs.interface.core.models.__common__ import (
    DataResponseModel,
    PaginationModel,
    ProcessModel,
    ProcessCreateRequest,
    ProcessUpdateRequest,
    optional_field_alt as optional_field,
    process_required_fields,
)
from lqs.interface.core import models


class Digestion(ProcessModel["Digestion"]):
    log_id: UUID = Field(
        ..., description="The ID of the log to which the digestion belongs."
    )
    group_id: UUID = Field(
        ..., description="The ID of the group to which the digestion belongs."
    )
    name: Optional[str] = Field(
        None, description="The name of the digestion for reference (not unique)."
    )
    note: Optional[str] = Field(
        None, description="A general note about the digestion for reference."
    )
    context: Optional[dict] = Field(
        None, description="Arbitrary JSON context for the digestion."
    )

    model_config = ConfigDict(
        json_schema_extra=lambda x: x["required"].extend(
            ["name", "note", "context"] + process_required_fields
        )
    )

    def list_digestion_topics(
        self, threaded=False, **kwargs
    ) -> List["models.DigestionTopic"]:
        return self._list_all_subresources(
            list_method=self._app.list.digestion_topic,
            threaded=threaded,
            digestion_id=self.id,
            **kwargs,
        )

    def list_digestion_parts(
        self, threaded=False, **kwargs
    ) -> List["models.DigestionPart"]:
        return self._list_all_subresources(
            list_method=self._app.list.digestion_part,
            threaded=threaded,
            digestion_id=self.id,
            **kwargs,
        )

    def fetch_log(self) -> "models.Log":
        return self._app.fetch.log(self.log_id).data

    def fetch_group(self) -> "models.Group":
        return self._app.fetch.group(self.group_id).data

    def fetch_workflow(self) -> Optional["models.Workflow"]:
        if self.workflow_id is None:
            return None
        return self._app.fetch.workflow(self.workflow_id).data


class DigestionDataResponse(DataResponseModel[Digestion]):
    pass


class DigestionListResponse(PaginationModel[Digestion]):
    pass


class DigestionCreateRequest(ProcessCreateRequest):
    log_id: UUID = Field(
        ..., description="The ID of the log to which the digestion should be added."
    )
    name: Optional[str] = Field(
        None, description="The name of the digestion (not unique)."
    )
    note: Optional[str] = Field(
        None, description="A general note about the digestion for reference."
    )
    context: Optional[dict] = Field(
        None, description="Arbitrary JSON context for the digestion."
    )


class DigestionUpdateRequest(ProcessUpdateRequest):
    name: Optional[str] = optional_field("The name of the digestion (not unique).")
    note: Optional[str] = optional_field(
        "A general note about the digestion for reference."
    )
    context: Optional[dict] = optional_field(
        "Arbitrary JSON context for the digestion."
    )



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from lqs.client import RESTClient
from lqs.client.admin.create import Create


class Admin:
    def __init__(self, app: "RESTClient"):
        self.app = app

        self.create = Create(app=app)



from typing import Optional
from uuid import UUID

from pydantic import BaseModel, Field, ConfigDict

from lqs.interface.core.models.__common__ import (
    DataResponseModel,
    CommonModel,
    PaginationModel,
    optional_field,
    Int64,
)
from lqs.interface.core import models


class DigestionTopic(CommonModel):
    digestion_id: UUID = Field(
        ...,
        description="The ID of the associated digestion.",
    )
    topic_id: UUID = Field(
        ...,
        description="The ID of the associated topic.",
    )
    start_time: Optional[Int64] = Field(
        None,
        description="The timestamp of the first record to be digested.",
    )
    end_time: Optional[Int64] = Field(
        None,
        description="The timestamp of the last record to be digested.",
    )
    frequency: Optional[float] = Field(
        None,
        description="The frequency of the records to be digested.",
    )
    query_data_filter: Optional[dict] = Field(
        None,
        description="A JSON filter applied to the record's query data for the digestion.",
    )
    context_filter: Optional[dict] = Field(
        None,
        description="A JSON filter applied to the record's context for the digestion.",
    )

    model_config = ConfigDict(
        json_schema_extra=lambda x: x["required"].extend(
            [
                "start_time",
                "end_time",
                "frequency",
                "query_data_filter",
                "context_filter",
            ]
        )
    )

    def fetch_digestion(self) -> "models.Digestion":
        return self._app.fetch.digestion(self.digestion_id).data

    def fetch_topic(self) -> "models.Topic":
        return self._app.fetch.topic(self.topic_id).data


class DigestionTopicDataResponse(DataResponseModel[DigestionTopic]):
    pass


class DigestionTopicListResponse(PaginationModel[DigestionTopic]):
    pass


class DigestionTopicForDigestionCreateRequest(BaseModel):
    topic_id: UUID
    start_time: Optional[Int64]
    end_time: Optional[Int64]
    frequency: Optional[float] = None
    query_data_filter: Optional[dict] = None
    context_filter: Optional[dict] = None


class DigestionTopicCreateRequest(DigestionTopicForDigestionCreateRequest):
    digestion_id: UUID


class DigestionTopicUpdateRequest(BaseModel):
    start_time: Optional[Int64] = optional_field
    end_time: Optional[Int64] = optional_field
    frequency: Optional[float] = optional_field
    query_data_filter: Optional[dict] = optional_field
    context_filter: Optional[dict] = optional_field



# Generated from RosMessageParser.g4 by ANTLR 4.13.0
from antlr4 import *

if "." in __name__:
    from .RosMessageParser import RosMessageParser
else:
    from RosMessageParser import RosMessageParser

# This class defines a complete listener for a parse tree produced by RosMessageParser.
class RosMessageParserListener(ParseTreeListener):

    # Enter a parse tree produced by RosMessageParser#ros_file_input.
    def enterRos_file_input(self, ctx: RosMessageParser.Ros_file_inputContext):
        pass

    # Exit a parse tree produced by RosMessageParser#ros_file_input.
    def exitRos_file_input(self, ctx: RosMessageParser.Ros_file_inputContext):
        pass

    # Enter a parse tree produced by RosMessageParser#ros_message_input.
    def enterRos_message_input(self, ctx: RosMessageParser.Ros_message_inputContext):
        pass

    # Exit a parse tree produced by RosMessageParser#ros_message_input.
    def exitRos_message_input(self, ctx: RosMessageParser.Ros_message_inputContext):
        pass

    # Enter a parse tree produced by RosMessageParser#ros_action_input.
    def enterRos_action_input(self, ctx: RosMessageParser.Ros_action_inputContext):
        pass

    # Exit a parse tree produced by RosMessageParser#ros_action_input.
    def exitRos_action_input(self, ctx: RosMessageParser.Ros_action_inputContext):
        pass

    # Enter a parse tree produced by RosMessageParser#ros_service_input.
    def enterRos_service_input(self, ctx: RosMessageParser.Ros_service_inputContext):
        pass

    # Exit a parse tree produced by RosMessageParser#ros_service_input.
    def exitRos_service_input(self, ctx: RosMessageParser.Ros_service_inputContext):
        pass

    # Enter a parse tree produced by RosMessageParser#rosbag_input.
    def enterRosbag_input(self, ctx: RosMessageParser.Rosbag_inputContext):
        pass

    # Exit a parse tree produced by RosMessageParser#rosbag_input.
    def exitRosbag_input(self, ctx: RosMessageParser.Rosbag_inputContext):
        pass

    # Enter a parse tree produced by RosMessageParser#rosbag_nested_message.
    def enterRosbag_nested_message(
        self, ctx: RosMessageParser.Rosbag_nested_messageContext
    ):
        pass

    # Exit a parse tree produced by RosMessageParser#rosbag_nested_message.
    def exitRosbag_nested_message(
        self, ctx: RosMessageParser.Rosbag_nested_messageContext
    ):
        pass

    # Enter a parse tree produced by RosMessageParser#ros_message.
    def enterRos_message(self, ctx: RosMessageParser.Ros_messageContext):
        pass

    # Exit a parse tree produced by RosMessageParser#ros_message.
    def exitRos_message(self, ctx: RosMessageParser.Ros_messageContext):
        pass

    # Enter a parse tree produced by RosMessageParser#field_declaration.
    def enterField_declaration(self, ctx: RosMessageParser.Field_declarationContext):
        pass

    # Exit a parse tree produced by RosMessageParser#field_declaration.
    def exitField_declaration(self, ctx: RosMessageParser.Field_declarationContext):
        pass

    # Enter a parse tree produced by RosMessageParser#constant_declaration.
    def enterConstant_declaration(
        self, ctx: RosMessageParser.Constant_declarationContext
    ):
        pass

    # Exit a parse tree produced by RosMessageParser#constant_declaration.
    def exitConstant_declaration(
        self, ctx: RosMessageParser.Constant_declarationContext
    ):
        pass

    # Enter a parse tree produced by RosMessageParser#comment.
    def enterComment(self, ctx: RosMessageParser.CommentContext):
        pass

    # Exit a parse tree produced by RosMessageParser#comment.
    def exitComment(self, ctx: RosMessageParser.CommentContext):
        pass

    # Enter a parse tree produced by RosMessageParser#identifier.
    def enterIdentifier(self, ctx: RosMessageParser.IdentifierContext):
        pass

    # Exit a parse tree produced by RosMessageParser#identifier.
    def exitIdentifier(self, ctx: RosMessageParser.IdentifierContext):
        pass

    # Enter a parse tree produced by RosMessageParser#type_.
    def enterType_(self, ctx: RosMessageParser.Type_Context):
        pass

    # Exit a parse tree produced by RosMessageParser#type_.
    def exitType_(self, ctx: RosMessageParser.Type_Context):
        pass

    # Enter a parse tree produced by RosMessageParser#ros_type.
    def enterRos_type(self, ctx: RosMessageParser.Ros_typeContext):
        pass

    # Exit a parse tree produced by RosMessageParser#ros_type.
    def exitRos_type(self, ctx: RosMessageParser.Ros_typeContext):
        pass

    # Enter a parse tree produced by RosMessageParser#array_type.
    def enterArray_type(self, ctx: RosMessageParser.Array_typeContext):
        pass

    # Exit a parse tree produced by RosMessageParser#array_type.
    def exitArray_type(self, ctx: RosMessageParser.Array_typeContext):
        pass

    # Enter a parse tree produced by RosMessageParser#variable_array_type.
    def enterVariable_array_type(
        self, ctx: RosMessageParser.Variable_array_typeContext
    ):
        pass

    # Exit a parse tree produced by RosMessageParser#variable_array_type.
    def exitVariable_array_type(self, ctx: RosMessageParser.Variable_array_typeContext):
        pass

    # Enter a parse tree produced by RosMessageParser#fixed_array_type.
    def enterFixed_array_type(self, ctx: RosMessageParser.Fixed_array_typeContext):
        pass

    # Exit a parse tree produced by RosMessageParser#fixed_array_type.
    def exitFixed_array_type(self, ctx: RosMessageParser.Fixed_array_typeContext):
        pass

    # Enter a parse tree produced by RosMessageParser#integral_type.
    def enterIntegral_type(self, ctx: RosMessageParser.Integral_typeContext):
        pass

    # Exit a parse tree produced by RosMessageParser#integral_type.
    def exitIntegral_type(self, ctx: RosMessageParser.Integral_typeContext):
        pass

    # Enter a parse tree produced by RosMessageParser#floating_point_type.
    def enterFloating_point_type(
        self, ctx: RosMessageParser.Floating_point_typeContext
    ):
        pass

    # Exit a parse tree produced by RosMessageParser#floating_point_type.
    def exitFloating_point_type(self, ctx: RosMessageParser.Floating_point_typeContext):
        pass

    # Enter a parse tree produced by RosMessageParser#temporal_type.
    def enterTemporal_type(self, ctx: RosMessageParser.Temporal_typeContext):
        pass

    # Exit a parse tree produced by RosMessageParser#temporal_type.
    def exitTemporal_type(self, ctx: RosMessageParser.Temporal_typeContext):
        pass

    # Enter a parse tree produced by RosMessageParser#string_type.
    def enterString_type(self, ctx: RosMessageParser.String_typeContext):
        pass

    # Exit a parse tree produced by RosMessageParser#string_type.
    def exitString_type(self, ctx: RosMessageParser.String_typeContext):
        pass

    # Enter a parse tree produced by RosMessageParser#boolean_type.
    def enterBoolean_type(self, ctx: RosMessageParser.Boolean_typeContext):
        pass

    # Exit a parse tree produced by RosMessageParser#boolean_type.
    def exitBoolean_type(self, ctx: RosMessageParser.Boolean_typeContext):
        pass

    # Enter a parse tree produced by RosMessageParser#sign.
    def enterSign(self, ctx: RosMessageParser.SignContext):
        pass

    # Exit a parse tree produced by RosMessageParser#sign.
    def exitSign(self, ctx: RosMessageParser.SignContext):
        pass

    # Enter a parse tree produced by RosMessageParser#integral_value.
    def enterIntegral_value(self, ctx: RosMessageParser.Integral_valueContext):
        pass

    # Exit a parse tree produced by RosMessageParser#integral_value.
    def exitIntegral_value(self, ctx: RosMessageParser.Integral_valueContext):
        pass

    # Enter a parse tree produced by RosMessageParser#floating_point_value.
    def enterFloating_point_value(
        self, ctx: RosMessageParser.Floating_point_valueContext
    ):
        pass

    # Exit a parse tree produced by RosMessageParser#floating_point_value.
    def exitFloating_point_value(
        self, ctx: RosMessageParser.Floating_point_valueContext
    ):
        pass

    # Enter a parse tree produced by RosMessageParser#bool_value.
    def enterBool_value(self, ctx: RosMessageParser.Bool_valueContext):
        pass

    # Exit a parse tree produced by RosMessageParser#bool_value.
    def exitBool_value(self, ctx: RosMessageParser.Bool_valueContext):
        pass

    # Enter a parse tree produced by RosMessageParser#string_value.
    def enterString_value(self, ctx: RosMessageParser.String_valueContext):
        pass

    # Exit a parse tree produced by RosMessageParser#string_value.
    def exitString_value(self, ctx: RosMessageParser.String_valueContext):
        pass


del RosMessageParser



from lqs.transcode.ros1.RosMessageParserVisitorForMD5 import (
    RosMessageParserVisitorForMD5,
)
from lqs.transcode.ros1.ros_deserialization_methods import *  # noqa  # used by the generated code we execute dynamically
from lqs.transcode.ros1.RosMessageLexer import RosMessageLexer
from lqs.transcode.ros1.RosMessageParser import RosMessageParser
from lqs.transcode.ros1.RosMessageParserVisitor import (
    RosMessageParserVisitor,
    RosMessageParserVisitorForJsonSchema,
)
from antlr4 import CommonTokenStream, InputStream


import sys
import types

# TODO consider using slots like genpy does

header_def = '''
# Standard metadata for higher-level stamped data types.
# This is generally used to communicate timestamped data 
# in a particular coordinate frame.
# 
# sequence ID: consecutively increasing ID 
uint32 seq
#Two-integer timestamp that is expressed as:
# * stamp.sec: seconds (stamp_secs) since epoch (in Python the variable is called 'secs')
# * stamp.nsec: nanoseconds since stamp_secs (in Python the variable is called 'nsecs')
# time-handling sugar is provided by the client library
time stamp
#Frame this data is associated with
string frame_id
'''

class RosMessageDeserializer:
    def __init__(self, trim_size=None, generate_unnested_schema=True) -> None:
        self._class_registry = {}
        self._trim_size = trim_size
        self.generate_unnested_schema = generate_unnested_schema
        self._register(
            "std_msgs/Header",header_def
        )

    def _register(self, message_type, message_type_data) -> None:
        lexer = RosMessageLexer(InputStream(message_type_data))
        stream = CommonTokenStream(lexer)
        parser = RosMessageParser(stream)
        tree = parser.rosbag_input()
        if parser.getNumberOfSyntaxErrors() > 0:
            raise SyntaxError("syntax errors in ROS message definition")
        else:
            vinterp = RosMessageParserVisitor(message_class_name=message_type)
            class_registry = vinterp.visit(tree)
            if class_registry is None:
                raise ValueError("class_registry is None")
            class_mapping = {}
            for canonical_name, python_class_defintion in class_registry.items():
                package, class_name = canonical_name.rsplit(".")
                exec(
                    python_class_defintion + f"\n{class_name} = {class_name}",
                    None,
                    class_mapping,
                )
                if package not in sys.modules:
                    temp_package = types.ModuleType(
                        package, "This is a fake module for " + package
                    )
                    sys.modules[package] = temp_package
                else:
                    temp_package = sys.modules[package]
                class_mapping[class_name].__module__ = package
                temp_package.__dict__.update(
                    {
                        class_name: class_mapping[class_name],
                    }
                )
            self._class_registry.update(class_registry)

    def deserialize(self, message_bytes, message_type, message_type_data=None):
        main_package, main_class_name = message_type.rsplit("/")
        if message_type.replace("/", ".") not in self._class_registry:
            self._register(message_type, message_type_data)
        main_cls = getattr(__import__(main_package), main_class_name)
        res = main_cls()
        res.deserialize(message_bytes, TRIM_SIZE=self._trim_size)
        return res

    def get_json_schema(self, message_type, message_type_data) -> dict:
        lexer = RosMessageLexer(InputStream(message_type_data))
        stream = CommonTokenStream(lexer)
        parser = RosMessageParser(stream)
        tree = parser.rosbag_input()
        if parser.getNumberOfSyntaxErrors() > 0:
            raise SyntaxError("syntax errors in ROS message definition")
        else:
            vinterp = RosMessageParserVisitorForJsonSchema(
                message_class_name=message_type,
                generate_unnested_schema=self.generate_unnested_schema,
            )
            json_schema = vinterp.visit(tree)
            assert isinstance(json_schema, dict)
            return json_schema

    def get_md5(self, message_type, message_type_data):
        lexer = RosMessageLexer(InputStream(message_type_data))
        stream = CommonTokenStream(lexer)
        parser = RosMessageParser(stream)
        tree = parser.rosbag_input()
        if parser.getNumberOfSyntaxErrors() > 0:
            raise SyntaxError("syntax errors in ROS message definition")
        else:
            vinterp = RosMessageParserVisitorForMD5(
                message_class_name=message_type,
            )
            md5 = vinterp.visit(tree)
            return md5



from typing import List, Optional
from enum import Enum

from pydantic import BaseModel, ConfigDict

from lqs.interface.base.models.__common__ import (
    CommonModel,
    DataResponseModel,
    PaginationModel,
    optional_field,
)
from lqs.interface.base import models


class StatementEffect(str, Enum):
    allow = "allow"
    deny = "deny"


class StatementAction(str, Enum):
    read = "read"
    write = "write"
    wildcard = "*"

    create = "create"
    list = "list"
    fetch = "fetch"
    update = "update"
    delete = "delete"


class Statement(BaseModel):
    effect: StatementEffect
    action: List[StatementAction] = []
    resource: List[str] = []
    filter: Optional[dict] = {}

    model_config = ConfigDict(
        json_schema_extra={"required": ["effect", "action", "resource", "filter"]}
    )


class Policy(BaseModel):
    statement: List[Statement]


class Role(CommonModel):
    name: str
    policy: Policy
    note: Optional[str]

    disabled: bool
    default: bool
    managed: bool

    def list_users(self, threaded=False, **kwargs) -> list["models.User"]:
        return self._list_all_subresources(
            list_method=self._app.list.user,
            threaded=threaded,
            role_id=self.id,
            **kwargs,
        )


class RoleDataResponse(DataResponseModel[Role]):
    pass


class RoleListResponse(PaginationModel[Role]):
    pass


class RoleCreateRequest(BaseModel):
    name: str
    policy: Policy

    note: Optional[str] = None
    disabled: bool = False
    managed: bool = False
    default: bool = False


class RoleUpdateRequest(BaseModel):
    name: str = optional_field
    policy: Policy = optional_field
    note: Optional[str] = optional_field
    disabled: bool = optional_field
    default: bool = optional_field



import io
import json
from typing import List, Optional, Union

from pydantic import BaseModel
from lqs.interface.base.models import Int64


class JSONLHeader(BaseModel):
    is_header: bool

    log_name: str
    log_note: Optional[str] = None
    log_context: Optional[dict] = None

    group_name: str
    group_note: Optional[str] = None
    group_context: Optional[dict] = None

    extra: Optional[dict] = None


class JSONLTopicRow(BaseModel):
    is_topic_row: bool
    is_default_topic: bool

    topic_name: str
    topic_type_name: Optional[str] = None
    topic_type_schema: Optional[dict] = None
    topic_note: Optional[str] = None
    topic_context: Optional[dict] = None
    associated_topic_name: Optional[str] = None
    extra: Optional[dict] = None


class JSONLRecordRow(BaseModel):
    timestamp: Int64
    data: Union[dict, list]
    topic_name: Optional[str] = None
    extra: Optional[dict] = None


class JSONLLog(BaseModel):
    name: Optional[str] = None
    header: Optional[JSONLHeader] = None
    rows: List[Union[JSONLTopicRow, JSONLRecordRow]] = []

    def write_jsonl_file_like(self, f=None):
        if f is None:
            f = io.StringIO()
        if self.header:
            f.write(self.header.model_dump_json(exclude_unset=True))
            f.write("\n")
        for line in self.rows:
            if isinstance(line, dict):
                # write this as compressed as possible
                f.write(json.dumps(line, separators=(",", ":")))
            else:
                f.write(line.model_dump_json(exclude_unset=True))
            f.write("\n")
        f.seek(0)
        return f

    def write_jsonl_file(self, file_path=None):
        if file_path is None:
            if self.name is None:
                raise Exception("No file path provided and no name set.")
            file_path = f"{self.name}.jsonl"

        with open(file_path, "w") as f:
            if self.header:
                f.write(self.header.model_dump_json(exclude_unset=True))
                f.write("\n")
            for line in self.rows:
                if isinstance(line, dict):
                    f.write(json.dumps(line, separators=(",", ":")))
                else:
                    f.write(line.model_dump_json(exclude_unset=True))
                f.write("\n")

    def get_jsonl_data(self):
        lines = []
        if self.header:
            lines.append(self.header.model_dump_json(exclude_unset=True))
        for line in self.rows:
            lines.append(line.model_dump_json(exclude_unset=True))
        return "\n".join(lines).encode("utf-8")

    def load_jsonl_file(self, file_path: str, name: Optional[str] = None):
        self.name = name
        self.header = None
        self.rows = []
        if name is None:
            name = file_path.split("/")[-1].split(".")[0]
        with open(file_path, "r") as f:
            lines = f.readlines()
            for idx, line in enumerate(lines):
                data = json.loads(line)
                if idx == 0 and data.get("is_header"):
                    self.header = JSONLHeader(**data)
                elif data.get("is_topic_row"):
                    self.rows.append(JSONLTopicRow(**data))
                else:
                    self.rows.append(JSONLRecordRow(**data))
        return self



from abc import abstractmethod
from typing import Optional
from uuid import UUID

from lqs.interface.base.delete import DeleteInterface as BaseDeleteInterface
import lqs.interface.core.models as models


class DeleteInterface(BaseDeleteInterface):
    @abstractmethod
    def _callback(self, **kwargs):
        pass

    def callback(self, callback_id: UUID):
        return self._callback(
            callback_id=callback_id,
        )

    @abstractmethod
    def _digestion(self, **kwargs):
        pass

    def digestion(self, digestion_id: UUID, lock_token: Optional[str] = None):
        """
        Delete a digestion.

        Args:
            digestion_id (UUID): ID of the digestion to delete.
        Returns:
            None
        """
        return self._digestion(
            digestion_id=digestion_id,
            lock_token=lock_token,
        )

    @abstractmethod
    def _digestion_part(self, **kwargs):
        pass

    def digestion_part(
        self,
        digestion_part_id: UUID,
        digestion_id: Optional[UUID] = None,
        lock_token: Optional[str] = None,
    ):
        """
        Delete a digestion part.

        Args:
            digestion_part_id (UUID): ID of the digestion part to delete.
            digestion_id (UUID, optional): ID of the digestion to which the digestion part belongs.
        Returns:
            None
        """
        return self._digestion_part(
            digestion_part_id=digestion_part_id,
            digestion_id=digestion_id,
            lock_token=lock_token,
        )

    @abstractmethod
    def _digestion_topic(self, **kwargs):
        pass

    def digestion_topic(
        self,
        digestion_topic_id: UUID,
        digestion_id: Optional[UUID] = None,
        lock_token: Optional[str] = None,
    ):
        """
        Delete a digestion topic.

        Args:
            digestion_topic_id (UUID): ID of the digestion topic to delete.
            digestion_id (UUID, optional): ID of the digestion to which the digestion topic belongs.
        Returns:
            None
        """
        return self._digestion_topic(
            digestion_topic_id=digestion_topic_id,
            digestion_id=digestion_id,
            lock_token=lock_token,
        )

    @abstractmethod
    def _group(self, **kwargs):
        pass

    def group(self, group_id: UUID, lock_token: Optional[str] = None):
        """
        Delete a group.

        Args:
            group_id (UUID): ID of the group to delete.
        Returns:
            None
        """
        return self._group(
            group_id=group_id,
            lock_token=lock_token,
        )

    @abstractmethod
    def _hook(self, **kwargs):
        pass

    def hook(self, hook_id: UUID, workflow_id: Optional[UUID] = None):
        """
        Delete a hook.

        Args:
            hook_id (UUID): ID of the hook to delete.
            workflow_id (UUID, optional): ID of the workflow to which the hook belongs.
        Returns:
            None
        """
        return self._hook(
            hook_id=hook_id,
            workflow_id=workflow_id,
        )

    @abstractmethod
    def _ingestion(self, **kwargs):
        pass

    def ingestion(self, ingestion_id: UUID, lock_token: Optional[str] = None):
        """
        Delete an ingestion.

        Args:
            ingestion_id (UUID): ID of the ingestion to delete.
        Returns:
            None
        """
        return self._ingestion(
            ingestion_id=ingestion_id,
            lock_token=lock_token,
        )

    @abstractmethod
    def _ingestion_part(self, **kwargs):
        pass

    def ingestion_part(
        self,
        ingestion_part_id: UUID,
        ingestion_id: Optional[UUID],
        lock_token: Optional[str] = None,
    ):
        """
        Delete an ingestion part.

        Args:
            ingestion_part_id (UUID): ID of the ingestion part to delete.
            ingestion_id (UUID, optional): ID of the ingestion to which the ingestion part belongs.
        Returns:
            None
        """
        return self._ingestion_part(
            ingestion_part_id=ingestion_part_id,
            ingestion_id=ingestion_id,
            lock_token=lock_token,
        )

    @abstractmethod
    def _label(self, **kwargs):
        pass

    def label(self, label_id: UUID):
        """
        Delete a label.

        Args:
            label_id (UUID): ID of the label to delete.
        Returns:
            None
        """
        return self._label(
            label_id=label_id,
        )

    @abstractmethod
    def _log(self, **kwargs):
        pass

    def log(self, log_id: UUID, lock_token: Optional[str] = None):
        """
        Delete a log.

        Args:
            log_id (UUID): ID of the log to delete.
        Returns:
            None
        """
        return self._log(log_id=log_id, lock_token=lock_token)

    @abstractmethod
    def _log_object(self, **kwargs):
        pass

    def log_object(
        self, log_id: UUID, object_key: str, lock_token: Optional[str] = None
    ):
        """
        Delete a log object.

        Args:
            log_id (UUID): ID of the log to which the log object belongs.
            object_key (str): Key of the log object to delete.
        Returns:
            None
        """
        return self._log_object(
            log_id=log_id,
            object_key=object_key,
            lock_token=lock_token,
        )

    @abstractmethod
    def _object(self, **kwargs):
        pass

    def object(self, object_store_id: UUID, object_key: str):
        """
        Delete an object.

        Args:
            object_store_id (UUID): ID of the object store to which the object belongs.
            object_key (str): Key of the object to delete.
        Returns:
            None
        """
        return self._object(
            object_store_id=object_store_id,
            object_key=object_key,
        )

    @abstractmethod
    def _object_store(self, **kwargs):
        pass

    def object_store(self, object_store_id: UUID):
        """
        Delete an object store.

        Args:
            object_store_id (UUID): ID of the object store to delete.
        Returns:
            None
        """
        return self._object_store(
            object_store_id=object_store_id,
        )

    @abstractmethod
    def _query(self, **kwargs):
        pass

    def query(
        self,
        query_id: UUID,
        log_id: Optional[UUID] = None,
        lock_token: Optional[str] = None,
    ):
        """
        Delete a query.

        Args:
            query_id (UUID): ID of the query to delete.
            log_id (UUID, optional): ID of the log to which the query belongs.
        Returns:
            None
        """
        return self._query(
            query_id=query_id,
            log_id=log_id,
            lock_token=lock_token,
        )

    @abstractmethod
    def _record(self, **kwargs):
        pass

    def record(
        self,
        timestamp: models.Int64,
        topic_id: UUID,
        lock_token: Optional[str] = None,
    ):
        """
        Delete a record.

        Args:
            timestamp (int): Timestamp of the record to delete.
            topic_id (UUID): ID of the topic to which the record belongs.
        Returns:
            None
        """
        return self._record(
            timestamp=timestamp, topic_id=topic_id, lock_token=lock_token
        )

    @abstractmethod
    def _tag(self, **kwargs):
        pass

    def tag(
        self,
        tag_id: UUID,
        log_id: Optional[UUID] = None,
        lock_token: Optional[str] = None,
    ):
        """
        Delete a tag.

        Args:
            tag_id (UUID): ID of the tag to delete.
            log_id (UUID, optional): ID of the log to which the tag belongs.
        Returns:
            None
        """
        return self._tag(
            tag_id=tag_id,
            log_id=log_id,
            lock_token=lock_token,
        )

    @abstractmethod
    def _topic(self, **kwargs):
        pass

    def topic(self, topic_id: UUID, lock_token: Optional[str] = None):
        """
        Delete a topic.

        Args:
            topic_id (UUID): ID of the topic to delete.
        Returns:
            None
        """
        return self._topic(
            topic_id=topic_id,
            lock_token=lock_token,
        )

    @abstractmethod
    def _workflow(self, **kwargs):
        pass

    def workflow(self, workflow_id: UUID):
        """
        Delete a workflow.

        Args:
            workflow_id (UUID): ID of the workflow to delete.
        Returns:
            None
        """
        return self._workflow(
            workflow_id=workflow_id,
        )



from abc import ABC, abstractmethod
from uuid import UUID

import lqs.interface.base.models as models


class FetchInterface(ABC):
    @abstractmethod
    def _api_key(self, **kwargs) -> models.APIKeyDataResponse:
        pass

    def api_key(self, api_key_id: UUID) -> models.APIKeyDataResponse:
        return self._api_key(
            api_key_id=api_key_id,
        )

    @abstractmethod
    def _me(self, **kwargs) -> models.MeDataResponse:
        pass

    def me(self) -> models.MeDataResponse:
        return self._me()

    @abstractmethod
    def _role(self, **kwargs) -> models.RoleDataResponse:
        pass

    def role(self, role_id: UUID) -> models.RoleDataResponse:
        return self._role(
            role_id=role_id,
        )

    @abstractmethod
    def _user(self, **kwargs) -> models.UserDataResponse:
        pass

    def user(self, user_id: UUID) -> models.UserDataResponse:
        return self._user(
            user_id=user_id,
        )



from enum import Enum
from uuid import UUID
from typing import Optional
from datetime import datetime

from pydantic import BaseModel, Field

from lqs.interface.base.models.__common__ import (  # noqa: F401
    ResourceModel,
    CommonModel,
    T,
    TimeSeriesModel,
    DataResponseModel,
    PaginationModel,
    PatchOperation,
    JSONFilter,
    optional_field,
    optional_deprecated_field,
    optional_field_alt,
    Int64,
    SpecialDict,
)


class ProcessState(str, Enum):
    ready = "ready"
    queued = "queued"
    processing = "processing"
    finalizing = "finalizing"
    completed = "completed"
    failed = "failed"
    cancelled = "cancelled"
    archived = "archived"


class ProcessType(str, Enum):
    ingestion = "ingestion"
    ingestion_part = "ingestion_part"
    digestion = "digestion"
    digestion_part = "digestion_part"


class TypeEncoding(str, Enum):
    ros1 = "ros1"
    rbuf = "rbuf"
    json = "json"
    cdr = "cdr"


class LockModel(BaseModel):
    locked: bool = Field(
        ...,
        description="Whether the process is locked (i.e. cannot be modified).",
    )
    locked_by: Optional[UUID] = Field(
        ..., description="The ID of the user who locked the resource."
    )
    locked_at: Optional[datetime] = Field(
        ..., description="The time at which the resource was locked."
    )
    lock_token: Optional[str] = Field(
        ..., description="The token used to lock the resource."
    )


class LockCreateRequest(BaseModel):
    locked: bool = False


class LockUpdateRequest(BaseModel):
    locked: bool = optional_field


class ProcessModel(CommonModel[T], LockModel):
    # These fields are used by the user to control the process.
    workflow_id: Optional[UUID] = Field(
        None,
        description="The ID of the workflow to be executed during state transitions.",
    )
    workflow_context: Optional[dict] = Field(
        None,
        description="The context to be passed to the workflow during state transitions.",
    )
    state: ProcessState = Field(..., description="The current state of the process.")

    # These fields should not be modified by a user and only by the process.
    progress: Optional[float] = Field(
        None,
        description="The progress of the process for the current state (a float in [0,1]).",
    )
    previous_state: Optional[ProcessState] = Field(
        None,
        description="The previous state of the process.",
    )
    transitioned_at: Optional[datetime] = Field(
        None,
        description="The time at which the process transitioned to the current state.",
    )
    error: Optional[dict] = Field(
        None,
        description="The name of the error that occurred during the process.",
        deprecated="Use error_payload instead.",
    )

    error_name: Optional[str] = Field(
        None,
        description="The name of the error that occurred during the process.",
    )
    error_message: Optional[str] = Field(
        None,
        description="The message of the error that occurred during the process.",
    )
    error_payload: Optional[dict] = Field(
        None,
        description="The payload of the error that occurred during the process.",
    )


process_required_fields = [
    "workflow_id",
    "workflow_context",
    "progress",
    "previous_state",
    "transitioned_at",
    "error",
    "error_name",
    "error_message",
    "error_payload",
]


class ProcessCreateRequest(LockCreateRequest):
    workflow_id: Optional[UUID] = None
    workflow_context: Optional[dict] = None
    state: ProcessState = ProcessState.ready


class ProcessUpdateRequest(LockUpdateRequest):
    workflow_id: Optional[UUID] = optional_field
    workflow_context: Optional[dict] = optional_field
    state: ProcessState = optional_field
    progress: Optional[float] = optional_field
    error_name: Optional[str] = optional_field
    error_message: Optional[str] = optional_field
    error_payload: Optional[dict] = optional_field
    error: Optional[dict] = optional_deprecated_field



from typing import Optional

from pydantic import BaseModel, Field, ConfigDict

from lqs.interface.core.models.__common__ import (
    CommonModel,
    DataResponseModel,
    PaginationModel,
    optional_field,
)


class Callback(CommonModel):
    name: str = Field(..., description="The name of the callback for reference.")
    note: Optional[str] = Field(
        None, description="A general note about the callback for reference."
    )
    context: Optional[dict] = Field(
        None, description="Arbitrary JSON context for the callback."
    )
    managed: bool = Field(
        ..., description="Whether the callback is managed by the system."
    )
    default: bool = Field(
        ..., description="Whether the callback is the default callback."
    )
    disabled: bool = Field(
        ..., description="Whether the callback is disabled and cannot be used."
    )
    uri: Optional[str] = Field(
        None, description="The URI to which the callback should send requests."
    )
    parameter_schema: Optional[dict] = Field(
        None, description="The parameter JSON schema for the callback."
    )

    model_config = ConfigDict(
        json_schema_extra=lambda x: x["required"].extend(
            ["note", "context", "uri", "parameter_schema"]
        )
    )


class CallbackDataResponse(DataResponseModel[Callback]):
    pass


class CallbackListResponse(PaginationModel[Callback]):
    pass


class CallbackCreateRequest(BaseModel):
    name: str
    note: Optional[str] = None
    context: Optional[dict] = None
    managed: bool = False
    default: bool = False
    disabled: bool = False
    uri: Optional[str] = None
    secret: Optional[str] = None
    parameter_schema: Optional[dict] = None


class CallbackUpdateRequest(BaseModel):
    name: str = optional_field
    note: Optional[str] = optional_field
    context: Optional[dict] = optional_field
    managed: bool = optional_field
    default: bool = optional_field
    disabled: bool = optional_field
    uri: Optional[str] = optional_field
    secret: Optional[str] = optional_field
    parameter_schema: Optional[dict] = optional_field



import io
import os
import inspect
import logging
import random
import time
import base64
import struct
from functools import partial
from typing import Optional, Union

import av
import s3path
import bz2
import lz4.frame
import lz4.block
import zstandard
import numpy as np
import DracoPy as draco
from pythonjsonlogger import jsonlogger
from PIL import Image as ImagePIL
from PIL import ImageDraw
from mcap.data_stream import ReadDataStream
from mcap.records import Chunk


class ContextLoggerAdapter(logging.LoggerAdapter):
    def process(self, msg, kwargs):
        kwargs["extra"] = self.extra
        return msg, kwargs

    def log(self, level, msg, *args, **kwargs):
        """
        Override the default 'log' method to inject custom contextual information into the LogRecord.
        """
        if self.isEnabledFor(level):
            frame = inspect.currentframe()
            # the frame is two levels up
            stack_info = inspect.getouterframes(frame)[3][0]
            if stack_info:
                # Extracting filename, line number, and function name from the frame
                filename = stack_info.f_code.co_filename
                lineno = stack_info.f_lineno
                func_name = stack_info.f_code.co_name
                record = self.logger.makeRecord(
                    self.logger.name,
                    level,
                    filename,
                    lineno,
                    msg,
                    args,
                    None,
                    func_name,
                    extra=self.extra,
                )
                self.logger.handle(record)


def get_logger(
    name,
    level: Optional[str] = None,
    log_to_file: bool = False,
    json_logging: bool = False,
    correlation_id: str | None = None,
):
    log_level = level or os.environ.get("LOG_LEVEL", "INFO")
    if isinstance(log_level, str):
        log_level = log_level.upper()
    logger = logging.getLogger(name)
    logger.setLevel(log_level)
    # logger.propagate = False
    if logger.hasHandlers():
        return ContextLoggerAdapter(logger, {"correlation_id": correlation_id})
    correlation_id_str = "%(correlation_id)s:" if correlation_id else ""
    if json_logging:
        formatter = jsonlogger.JsonFormatter(
            f"%(asctime)s %(levelname)s {correlation_id_str} %(name)s %(filename)s %(funcName)s %(lineno)s %(message)s"
        )
        handler = logging.StreamHandler()
        handler.setFormatter(formatter)
        logger.addHandler(handler)
    else:
        formatter = logging.Formatter(
            f"%(asctime)s  (%(levelname)s - %(name)s): {correlation_id_str} %(message)s"
        )
        handler = logging.StreamHandler()
        handler.setFormatter(formatter)
        logger.addHandler(handler)

        if log_to_file:
            handler = logging.FileHandler("lqs.log")
            handler.setFormatter(formatter)
            logger.addHandler(handler)
    return ContextLoggerAdapter(logger, {"correlation_id": correlation_id})


def attempt_with_retries(
    func,
    fargs=None,
    fkwargs=None,
    exceptions=Exception,
    tries=3,
    delay=1.0,
    max_delay=None,
    backoff=2,
    jitter=0,
    message=None,
    logging_func=None,
):
    args = fargs if fargs else list()
    kwargs = fkwargs if fkwargs else dict()
    f = partial(func, *args, **kwargs)
    _tries, _delay = tries, delay
    attempt = 1
    while _tries:
        try:
            return f()
        except exceptions as e:
            _tries -= 1
            if not _tries:
                raise

            if logging_func is not None:
                if message is None:
                    warning_message = f"Failed on attempt {attempt}: {e}; retrying in {_delay} seconds..."
                else:
                    warning_message = f'Failed "{message}" on attempt {attempt}: {e}; retrying in {_delay} seconds...'
                logging_func(warning_message)

            attempt += 1
            time.sleep(_delay)
            _delay *= backoff
            if isinstance(jitter, tuple):
                _delay += random.uniform(*jitter)
            else:
                _delay += jitter
            if max_delay is not None:
                _delay = min(_delay, max_delay)


def get_relative_object_path(object_key: str, source: str):
    if source is None:
        return object_key
    if not object_key.endswith("/"):
        object_key = object_key.rsplit("/", 1)[0] + "/"
    return str(s3path.PureS3Path(object_key + source))


def decompress_chunk_bytes(chunk_bytes, chunk_compression, chunk_length=None):
    if chunk_compression == "bz2":
        decompressed_bytes = bz2.decompress(chunk_bytes)
    elif chunk_compression == "lz4":
        decompressed_bytes = lz4.frame.decompress(chunk_bytes)
    elif chunk_compression == "zstd_mcap":
        chunk = ModifiedChunk.read(ReadDataStream(io.BytesIO(chunk_bytes)))
        return zstandard.decompress(chunk.data, chunk.uncompressed_size)
    elif chunk_compression == "zstd":
        decompressed_bytes = zstandard.decompress(chunk_bytes)
    elif chunk_compression == "lz4_block":
        if chunk_length is None:
            raise ValueError("Must specify chunk_length when decompressing lz4_block")
        decompressed_bytes = lz4.block.decompress(chunk_bytes, chunk_length)
    elif chunk_compression in [None, "none", ""]:
        decompressed_bytes = chunk_bytes
    elif chunk_compression == "mp4":
        decompressed_bytes = chunk_bytes
    else:
        raise ValueError(f"Unknown chunk compression type: {chunk_compression}")
    return decompressed_bytes


# Image

image_type_names = [
    "sensor_msgs/Image",
    "sensor_msgs/CompressedImage",
    "stereo_msgs/DisparityImage",
    "sensor_msgs/msg/Image",
    "sensor_msgs/msg/CompressedImage",
    "stereo_msgs/msg/DisparityImage",
    "ark::image::Image",
    "logqs.image",
    "logqs.inference.depth-estimation",
    "logqs.inference.image-to-image",
    "lqs_cutter/navcon_image",
]


def image_to_base64(
    image: ImagePIL.Image, format: str = "WEBP", format_params: dict = {}
) -> str:
    buffered_img = io.BytesIO()
    image.save(buffered_img, format=format, **format_params)
    return base64.b64encode(buffered_img.getvalue()).decode("utf-8")


def get_complete_segmentation_image(output, meta):
    image_size = (
        meta["image"]["width"],
        meta["image"]["height"],
    )
    img = ImagePIL.new("RGB", image_size, (0, 0, 0))
    for segmentation in output:
        mask_data = base64.b64decode(segmentation["mask"])
        mask = ImagePIL.open(io.BytesIO(mask_data)).convert("L")

        label = segmentation.get("label")
        if label is None:
            color = (
                random.randint(0, 255),
                random.randint(0, 255),
                random.randint(0, 255),
            )
        else:
            color = int(hash(label) % 0xFFFFFF)
            color = (
                color & 0xFF,
                (color >> 8) & 0xFF,
                (color >> 16) & 0xFF,
            )
        draw = ImageDraw.Draw(img)
        draw.bitmap((0, 0), mask, fill=color)
    return img


def convert_avc1_to_annexb(data, filter_parameter_sets=True):
    # Wrap data in a memoryview and cache its length.
    mv = memoryview(data)
    mv_len = len(mv)
    offset = 0
    annexb = bytearray()
    start_code = b"\x00\x00\x00\x01"

    while offset + 4 <= mv_len:
        nal_length = struct.unpack_from(">I", mv, offset)[0]
        offset += 4
        # Validate NAL length.
        if nal_length <= 0 or offset + nal_length > mv_len:
            break
        # Avoid parameter set duplication.
        nal_unit = mv[offset : offset + nal_length].tobytes()
        offset += nal_length
        nal_type = nal_unit[0] & 0x1F
        if filter_parameter_sets and nal_type in (7, 8):
            continue
        annexb.extend(start_code)
        annexb.extend(nal_unit)
    return bytes(annexb)


def convert_hvc1_to_annexb(data):
    mv = memoryview(data)
    total_len = len(mv)
    offset = 0
    result = bytearray()
    start_code = b"\x00\x00\x00\x01"

    while offset + 4 <= total_len:
        nal_length = struct.unpack_from(">I", mv, offset)[0]
        offset += 4
        if nal_length <= 0 or offset + nal_length > total_len:
            break
        result.extend(start_code)
        result.extend(mv[offset : offset + nal_length])
        offset += nal_length
    return bytes(result)


def generate_image_from_frames(extradata, frames, frame_data, encoding):
    # Create the decoder and set extradata.
    codec = av.codec.CodecContext.create(encoding, "r")
    codec.extradata = extradata

    # Precompute start_offset and wrap frame_data in a memoryview.
    start_offset = frames[0][2]
    fd_mv = memoryview(frame_data)
    fd_len = len(fd_mv)
    converter = convert_hvc1_to_annexb if encoding == "hevc" else convert_avc1_to_annexb

    decoded_frames = None
    for frame in frames:
        # Compute slice boundaries once.
        offset_in_frame = frame[2] - start_offset
        size = frame[1]
        # Check bounds.
        if fd_len < offset_in_frame + size:
            break
        # Slice out the NAL unit data.
        nal_data = fd_mv[offset_in_frame : offset_in_frame + size]
        # Call the converter; if nal_data is already a memoryview, let the converter wrap it.
        frame_annexb = converter(nal_data)
        packet = av.Packet(frame_annexb)
        decoded_frames = codec.decode(packet)
    # We expect a single decoded frame.
    assert (
        decoded_frames is not None and len(decoded_frames) == 1
    ), f"Delta frame decoding failed: {len(decoded_frames) if decoded_frames else 0} frames decoded, expected 1"
    decoded_frame = decoded_frames[0]
    image = decoded_frame.to_image()
    return image


def get_record_image_lqs(record_data, **kwargs) -> ImagePIL.Image:
    if "frames" in kwargs:
        # if we've passed in frame information, then this is a video frame
        frames = kwargs["frames"]
        encoding = kwargs["encoding"]

        extradata = kwargs.get("extradata", None)
        if isinstance(extradata, str):
            extradata = base64.b64decode(extradata)

        frame_data = record_data
        if isinstance(frame_data, str):
            frame_data = base64.b64decode(frame_data)

        img = generate_image_from_frames(extradata, frames, frame_data, encoding)
    elif isinstance(record_data, str):
        # if the record data is a string, it is a base64 encoded image
        try:
            image_data = base64.b64decode(record_data)
        except Exception as e:
            raise Exception(f"Failed to decode record data: {e}")
        try:
            img = ImagePIL.open(io.BytesIO(image_data))
        except Exception as e:
            raise Exception(f"Failed to open image: {e}")
    else:
        meta = record_data.get("meta")
        if meta:
            pipeline = meta.get("pipeline")
            if pipeline:
                task = pipeline.get("task")
                if task is None:
                    raise Exception(
                        "Expected to find 'task' in record data's 'meta.pipeline' field."
                    )
                if task == "depth-estimation":
                    image_data = base64.b64decode(record_data["output"]["depth"])
                    img = ImagePIL.open(io.BytesIO(image_data))
                elif task == "image-segmentation":
                    img = get_complete_segmentation_image(
                        output=record_data["output"], meta=meta
                    )
                elif task == "image-to-image":
                    image_data = base64.b64decode(record_data["output"])
                    img = ImagePIL.open(io.BytesIO(image_data))
                else:
                    raise Exception(
                        f"Task '{task}' not supported for image generation."
                    )
            else:
                raise Exception(
                    "Expected to find 'pipeline' in record data's 'meta' field."
                )
        else:
            raise Exception("Expected to find 'meta' in record data.")
    return img


def normalize_disparity_image(
    image: ImagePIL.Image, min_clip=74, max_clip=2592, **kwargs
):
    pixels = np.asarray(image)
    pixels = np.clip(pixels, min_clip, max_clip)
    pixels = ((pixels - min_clip) / (max_clip - min_clip)) * 255.0
    return ImagePIL.fromarray(pixels).convert("L")


def get_record_image_ros(record_data, renormalize=True, **kwargs) -> ImagePIL.Image:
    img_modes = {
        "16UC1": "I;16",
        "mono8": "L",
        "mono16": "I;16",
        "32FC1": "F",
        "8UC1": "L",
        "8UC3": "RGB",
        "rgb8": "RGB",
        "bgr8": "RGB",
        "rgba8": "RGBA",
        "bgra8": "RGBA",
        "bayer_rggb": "L",
        "bayer_rggb8": "L",
        "bayer_gbrg": "L",
        "bayer_gbrg8": "L",
        "bayer_grbg": "L",
        "bayer_grbg8": "L",
        "bayer_bggr": "L",
        "bayer_bggr8": "L",
        "yuv422": "YCbCr",
        "yuv411": "YCbCr",
    }

    record_image_format = record_data.get("format", None)
    if record_image_format is None:
        encoding = record_data.get("encoding", None)
        if not encoding:
            return None

        mode = img_modes[encoding]
        img = ImagePIL.frombuffer(
            mode,
            (record_data["width"], record_data["height"]),
            bytes(record_data["data"]),
            "raw",
            mode,
            0,
            1,
        )
        if encoding == "bgr8":
            b, g, r = img.split()
            img = ImagePIL.merge("RGB", (r, g, b))
        elif encoding in ["mono16", "16UC1", "32FC1"] and renormalize:
            if kwargs.get("normalize_disparity"):
                img = normalize_disparity_image(img, **kwargs)
            pixels = np.asarray(img)
            pixel_max = np.max(pixels)
            pixel_range = pixel_max - np.min(pixels)
            if pixel_range == 0:
                pixels = np.zeros_like(pixels)
            else:
                pixels = ((pixels - np.min(pixels)) / pixel_range) * 255.0
            img = ImagePIL.fromarray(pixels)
            img = img.convert("L")
    else:
        if record_data["format"] == "h264":
            import av

            codec = av.CodecContext.create("h264", "r")
            packets = av.packet.Packet(bytes(record_data["data"]))
            img = codec.decode(packets)[0].to_image()
        else:
            img = ImagePIL.open(io.BytesIO(record_data["data"]))
    return img


def get_record_image_ark(record_data, renormalize=True, **kwargs) -> ImagePIL.Image:
    data = bytes(record_data["data"])
    image_format = record_data["data_format"]
    image_format_mapping = {
        42: "BayerRg10",
        44: "BayerRg12",
        43: "BayerRg16",
        41: "BayerRg8",
        31: "Bgr",
        33: "Bgra",
        51: "DepthZ16",
        21: "Grey",
        81: "H264",
        2: "Jpeg",
        1: "MotionJpeg",
        71: "Nv12",
        32: "Rgb",
        34: "Rgba",
        0: "Unset",
        27: "Uv8",
        26: "Yuv420",
        25: "Yuyv",
    }
    if image_format in image_format_mapping:
        image_format = image_format_mapping[image_format]
    else:
        raise NotImplementedError(f"Message type {image_format} not supported.")
    image_format = image_format.lower()

    if image_format == "depthz16":
        ark_img = np.frombuffer(data, dtype=np.float16)

        if renormalize:
            pixel_range = np.max(ark_img) - np.min(ark_img)
            if pixel_range != 0:
                ark_img = ((ark_img - np.min(ark_img)) / pixel_range) * 65535.0

        ark_img = ark_img.astype(np.uint16)
        data = ark_img
        image_mode = "I;16"
        img = ImagePIL.frombuffer(
            image_mode,
            (record_data["width"], record_data["height"]),
            data,
            "raw",
            image_mode,
            0,
            1,
        )
        pixels = np.asarray(img)
        pixel_range = np.max(pixels) - np.min(pixels)
        if pixel_range == 0:
            pixels = np.zeros_like(pixels)
        else:
            pixels = ((pixels - np.min(pixels)) / pixel_range) * 255.0
        img = ImagePIL.fromarray(pixels)
        img = img.convert("L")
    elif image_format == "grey":
        image_mode = "L"
        img = ImagePIL.frombuffer(
            image_mode,
            (record_data["width"], record_data["height"]),
            data,
            "raw",
            image_mode,
            0,
            1,
        )
    elif image_format == "rgb" or image_format == "bgr":
        image_mode = "RGB"
        img = ImagePIL.frombuffer(
            image_mode,
            (record_data["width"], record_data["height"]),
            data,
            "raw",
            image_mode,
            0,
            1,
        )
        if image_format == "bgr":
            b, g, r = img.split()
            img = ImagePIL.merge("RGB", (r, g, b))
    elif image_format == "rgba" or image_format == "bgra":
        image_mode = "RGBA"
        img = ImagePIL.frombuffer(
            image_mode,
            (record_data["width"], record_data["height"]),
            data,
            "raw",
            image_mode,
            0,
            1,
        )
        if image_format == "bgra":
            b, g, r, a = img.split()
            img = ImagePIL.merge("RGBA", (r, g, b, a))
    elif image_format == "jpeg":
        img = ImagePIL.open(io.BytesIO(bytes(record_data["data"])))
    else:
        if image_format == "h264":
            import av

            codec = av.CodecContext.create("h264", "r")
            packets = av.packet.Packet(record_data.data)
            img = codec.decode(packets)[0].to_image()
            # this may not work
        elif image_format.lower() == "nv12":
            width = record_data["width"]
            height = record_data["height"]
            data = record_data["data"]
            Y_size = width * height
            UV_size = (
                Y_size // 2
            )  # There are half as many UV samples as Y samples in NV12

            # Separate Y, U, and V values
            Y = np.array(data[:Y_size], dtype=np.uint8).reshape((height, width))
            UV = np.array(data[Y_size : Y_size + UV_size], dtype=np.uint8).reshape(
                (height // 2, width)
            )

            # Deinterleave U and V components
            U = UV[:, ::2]
            V = UV[:, 1::2]

            # Up-sample U and V components to match Y's dimension
            U_upsampled = np.repeat(np.repeat(U, 2, axis=0), 2, axis=1)
            V_upsampled = np.repeat(np.repeat(V, 2, axis=0), 2, axis=1)

            # Stack the YUV components together to form the YUV image
            YUV = np.stack((Y, U_upsampled, V_upsampled), axis=-1).astype(np.uint8)

            # Convert YUV to RGB
            rgb_image = ImagePIL.fromarray(YUV, "YCbCr").convert("RGB")

            # Save image as png
            img = rgb_image
        else:
            raise NotImplementedError(
                f"Message type {image_format} not supported for image conversion."
            )
    return img


def get_record_image_cutter(record_data, renormalize=True, **kwargs) -> ImagePIL.Image:
    image_format = record_data["type"]
    images_data = record_data["data"][0]
    if image_format != "JPG":
        raise NotImplementedError(
            f"Message type {image_format} not supported for image conversion."
        )
    img = ImagePIL.open(io.BytesIO(bytes(images_data)))
    return img


def get_record_image(
    record_data,
    max_size: Optional[int] = None,
    format: str = "WEBP",
    format_params: dict = {},
    renormalize: bool = True,
    resample=ImagePIL.Resampling.NEAREST,
    reset_position: bool = True,
    return_bytes: bool = False,
    **kwargs,
) -> Union[ImagePIL.Image, io.BytesIO, None]:
    if isinstance(record_data, bytes):
        img = get_record_image_lqs(
            record_data=record_data, renormalize=renormalize, **kwargs
        )
    else:
        _format = record_data.get("format", None)
        _encoding = record_data.get("encoding", None)
        _data_format = record_data.get("data_format", None)
        _image = record_data.get("image", None)
        if record_data.get("type") and record_data.get("cameras"):
            img = get_record_image_cutter(
                record_data=record_data, renormalize=renormalize, **kwargs
            )
        elif _encoding is not None:
            # ROS sensor_msgs/Image
            img = get_record_image_ros(
                record_data=record_data, renormalize=renormalize, **kwargs
            )
        elif _format is not None:
            # ROS sensor_msgs/CompressedImage
            img = get_record_image_ros(
                record_data=record_data, renormalize=renormalize, **kwargs
            )
        elif _data_format is not None:
            # ARK ark::image::Image
            img = get_record_image_ark(
                record_data=record_data, renormalize=renormalize, **kwargs
            )
        elif _image is not None:
            # ROS stereo_msgs/DisparityImage
            img = get_record_image_ros(
                record_data=_image, renormalize=renormalize, **kwargs
            )
        else:
            # we default to LQS image formats
            img = get_record_image_lqs(
                record_data=record_data, renormalize=renormalize, **kwargs
            )

    if max_size is not None:
        if img.width > max_size or img.height > max_size:
            img.thumbnail((max_size, max_size), resample=resample)

    if return_bytes:
        buffered_img = io.BytesIO()
        img.save(buffered_img, format=format, **format_params)
        if reset_position:
            buffered_img.seek(0)
        return buffered_img
    else:
        return img


# Point Cloud

point_cloud_type_names = [
    "sensor_msgs/PointCloud2",
    "sensor_msgs/msg/PointCloud2",
]


def jet_colormap(values, normalize=True):
    if normalize:
        values = (values - np.min(values)) / (np.max(values) - np.min(values))
    four_values = 4 * values
    red = np.minimum(four_values - 1.5, -four_values + 4.5)
    green = np.minimum(four_values - 0.5, -four_values + 3.5)
    blue = np.minimum(four_values + 0.5, -four_values + 2.5)
    colors = np.stack(
        [
            255 * np.clip(red, 0, 1),
            255 * np.clip(green, 0, 1),
            255 * np.clip(blue, 0, 1),
        ],
        axis=-1,
    )
    return colors.astype(np.uint8)


def voxel_downsample(positions, colors, max_points, voxel_size=1.0):
    # Calculate voxel indices for each point
    voxel_indices = np.floor(positions / voxel_size).astype(int)

    # Create a structured array to associate positions with their voxel
    dtype = [("x", int), ("y", int), ("z", int)]
    structured_voxel_indices = np.array(
        list(zip(voxel_indices[:, 0], voxel_indices[:, 1], voxel_indices[:, 2])),
        dtype=dtype,
    )

    # Find unique voxels and the indices of points in those voxels
    unique_voxels, inverse_indices, voxel_counts = np.unique(
        structured_voxel_indices, return_inverse=True, return_counts=True, axis=0
    )

    # Calculate sampling probabilities inversely proportional to voxel density
    sampling_probabilities = 1.0 / voxel_counts[inverse_indices]
    sampling_probabilities /= sampling_probabilities.sum()

    # Sample points based on calculated probabilities
    selected_indices = np.random.choice(
        np.arange(len(positions)),
        size=min(max_points, len(positions)),
        replace=False,
        p=sampling_probabilities,
    )

    return (
        positions[selected_indices],
        colors[selected_indices] if colors is not None else None,
    )


def process_pointcloud_record_data(
    record_data,
    max_points=None,
    random_sample=True,
    voxel_size=10.0,
    use_colormap=False,
    max_radius=None,
    **kwargs,
):
    fields = record_data["fields"]
    valid_fields = ["x", "y", "z", "rgb", "intensity"]
    fields = [f for f in record_data["fields"] if f["name"] in valid_fields]
    if len(fields) == 0:
        if len(record_data["data"]) > 0:
            raise ValueError(
                "No fields found in point cloud data with non-zero data length."
            )
        return [], []

    assert [f["name"] for f in fields[:3]] == [
        "x",
        "y",
        "z",
    ], f"Expected fields to be ['x', 'y', 'z'], but got {fields[:3]}"

    point_cloud_data = bytes(record_data["data"])
    dtype_map = {
        1: np.int8,
        2: np.uint8,
        3: np.int16,
        4: np.uint16,
        5: np.int32,
        6: np.uint32,
        7: np.float32,
        8: np.float64,
    }
    dtype_list = []
    current_offset = 0
    for field in fields:
        field_name = field["name"]
        field_type = dtype_map[field["datatype"]]
        field_offset = field["offset"]
        # Check if there is padding needed
        if field_offset > current_offset:
            padding_size = field_offset - current_offset
            dtype_list.append(("padding" + str(current_offset), np.uint8, padding_size))
        dtype_list.append((field_name, field_type))
        current_offset = field_offset + np.dtype(field_type).itemsize

    if current_offset < record_data["point_step"]:
        padding_size = record_data["point_step"] - current_offset
        dtype_list.append(("padding" + str(current_offset), np.uint8, padding_size))

    dtype = np.dtype(dtype_list)
    all_data = np.frombuffer(point_cloud_data, dtype=dtype)

    # Access positions directly via field names, assuming the first three fields are x, y, z
    positions = np.vstack([all_data["x"], all_data["y"], all_data["z"]]).T

    colors = None
    if len(fields) > 3 and fields[3]["name"] == "rgb":
        # Extract and process RGB colors directly from the structured array
        rgb_ints = all_data["rgb"].view(
            np.uint32
        )  # Assuming 'rgb' is stored as floats and can be viewed as uint32
        red = (rgb_ints >> 16) & 255
        green = (rgb_ints >> 8) & 255
        blue = rgb_ints & 255
        colors = np.vstack([red, green, blue]).T.astype(np.uint8)
    elif len(fields) > 3 and fields[3]["name"] == "intensity":
        # Process intensity data if needed
        colors = all_data["intensity"].astype(
            np.uint8
        )  # Assuming 'intensity' directly gives the desired data
        if use_colormap:
            colors = jet_colormap(colors)
        else:
            colors = np.stack([colors] * 3, axis=-1)

    if max_radius is not None:
        indices = np.linalg.norm(positions, axis=1) < max_radius
        positions = positions[indices]
        if colors is not None:
            colors = colors[indices]

    if max_points and len(positions) > max_points:
        if random_sample:
            indices = np.random.choice(len(positions), max_points, replace=False)
            positions = positions[indices]
            if colors is not None:
                colors = colors[indices]
        else:
            positions, colors = voxel_downsample(
                positions, colors, max_points, voxel_size=voxel_size
            )

    return positions, colors


def encode_to_draco(
    positions,
    quant_bits=11,
    compression=7,
    remove_dups=True,
    colors=None,
    max_points=None,
):
    if len(positions) == 0:
        raise ValueError("Empty positions array for encoding to Draco.")

    draco_binary = draco.encode(
        positions,
        faces=None,
        quantization_bits=quant_bits,
        compression_level=0,
        quantization_range=-1,
        quantization_origin=None,
        create_metadata=False,
        preserve_order=False,
        colors=None if remove_dups else colors,
        tex_coord=None,
    )
    if remove_dups:
        _draco_points = draco.decode(draco_binary).points
        _, unique_indices = np.unique(_draco_points, axis=0, return_index=True)
        if max_points is not None and len(unique_indices) > max_points:
            unique_indices = np.random.choice(unique_indices, max_points, replace=False)

        _draco_points = _draco_points[unique_indices]
        if colors is not None:
            colors = colors[unique_indices]

        draco_binary = draco.encode(
            _draco_points,
            faces=None,
            quantization_bits=quant_bits,
            compression_level=compression,
            quantization_range=-1,
            quantization_origin=None,
            create_metadata=False,
            preserve_order=False,
            colors=colors,
            tex_coord=None,
        )
    return draco_binary


def decode_draco(draco_binary):
    return draco.decode(draco_binary).points, draco.decode(draco_binary).colors


# MCAP


class ModifiedChunk(Chunk):
    _chunk_start_offset: int
    _chunk_length: int

    def __init__(self, **kwargs):
        self._chunk_start_offset = kwargs.pop("_chunk_start_offset")
        self._chunk_length = kwargs.pop("_chunk_length")
        super().__init__(**kwargs)

    @staticmethod
    def read(stream: ReadDataStream):
        _chunk_start_offset = stream.count
        message_start_time = stream.read8()
        message_end_time = stream.read8()
        uncompressed_size = stream.read8()
        uncompressed_crc = stream.read4()
        compression_length = stream.read4()
        compression = str(stream.read(compression_length), "utf-8")
        data_length = stream.read8()
        if compression != "zstd":
            _chunk_start_offset = stream.count
            _chunk_length = data_length
        else:
            # we need to modify the chunk_length
            _chunk_length = (stream.count - _chunk_start_offset) + data_length
        data = stream.read(data_length)
        return ModifiedChunk(
            compression=compression,
            data=data,
            message_end_time=message_end_time,
            message_start_time=message_start_time,
            uncompressed_crc=uncompressed_crc,
            uncompressed_size=uncompressed_size,
            _chunk_start_offset=_chunk_start_offset,
            _chunk_length=_chunk_length,
        )



# make pylance ignore this file
# type: ignore
# Generated from RosMessageParser.g4 by ANTLR 4.13.0
import io
from typing import Union

from antlr4 import *

if "." in __name__:
    from .RosMessageParser import RosMessageParser
else:
    from RosMessageParser import RosMessageParser

# This class defines a complete generic visitor for a parse tree produced by RosMessageParser.


class RosMessageParserVisitor(ParseTreeVisitor):
    def __init__(self, message_class_name=""):
        self.message_class_name = message_class_name

    # Visit a parse tree produced by RosMessageParser#ros_file_input.
    def visitRos_file_input(self, ctx: RosMessageParser.Ros_file_inputContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by RosMessageParser#ros_message_input.
    def visitRos_message_input(self, ctx: RosMessageParser.Ros_message_inputContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by RosMessageParser#ros_action_input.
    def visitRos_action_input(self, ctx: RosMessageParser.Ros_action_inputContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by RosMessageParser#ros_service_input.
    def visitRos_service_input(self, ctx: RosMessageParser.Ros_service_inputContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by RosMessageParser#rosbag_input.
    def visitRosbag_input(self, ctx: RosMessageParser.Rosbag_inputContext):
        self.class_registry = {}
        res = self.visitChildren(ctx)
        return {
            k.replace("/", "."): v.getvalue() for k, v in self.class_registry.items()
        }

    # Visit a parse tree produced by RosMessageParser#rosbag_nested_message.
    def visitRosbag_nested_message(
        self, ctx: RosMessageParser.Rosbag_nested_messageContext
    ):
        self.message_class_name = ctx.getChild(1).getText()
        return self.visitChildren(ctx)

    # Visit a parse tree produced by RosMessageParser#ros_message.
    def visitRos_message(self, ctx: RosMessageParser.Ros_messageContext):
        self.class_registry[self.message_class_name] = io.StringIO()
        package, class_name = self.message_class_name.rsplit("/")
        # self.class_registry[self.message_class_name].write(f"from ros_serialization_methods import *\n")
        self.class_registry[self.message_class_name].write(f"class {class_name}:")
        self.class_registry[self.message_class_name].write(f'\n\tpackage = "{package}"')
        # collect all the non-comment children
        constant_declaration_list = list(
            ctx.getChildren(
                lambda x: isinstance(x, RosMessageParser.Constant_declarationContext)
            )
        )
        member_declaration_list = list(
            ctx.getChildren(
                lambda x: isinstance(x, RosMessageParser.Field_declarationContext)
            )
        )
        declared_members = [
            (member.getChild(1).getText(), member.getChild(0).getText())
            for member in member_declaration_list
            if isinstance(member, RosMessageParser.Field_declarationContext)
        ]
        # print out __slots__
        self.class_registry[self.message_class_name].write(
            f"\n\t__slots__ = ["
            + ",".join([f"'{member[0]}'" for member in declared_members])
            + "]"
        )
        # print out the _slot_types
        self.class_registry[self.message_class_name].write(
            f"\n\t_slot_types = ["
            + ",".join([f"'{member[1]}'" for member in declared_members])
            + "]"
        )

        self.class_registry[self.message_class_name].write("\n\tdef __init__(self):")

        for member in member_declaration_list:
            if isinstance(member, RosMessageParser.Field_declarationContext):
                # self.class_registry[self.message_class_name].write(f"\n\t\t{member.getChild(0).getText()}")
                self.class_registry[self.message_class_name].write(
                    f"\n\t\tself.{member.getChild(1).getText()} = None"
                )
            elif isinstance(member, RosMessageParser.Constant_declarationContext):
                # self.class_registry[self.message_class_name].write(f"\n\t\t{member.getChild(0).getText()}")
                self.class_registry[self.message_class_name].write(
                    f"\n\t\tself.{member.getChild(1).getText()} = {member.getChild(3).getText()}"
                )
            else:
                raise NotImplementedError(f"Unknown member type: {member.__class__}")

        self.class_registry[self.message_class_name].write(
            "\n\tdef __str__(self):\n\t\t return f\"{self.__class__.__name__}({', '.join([f'{k}={getattr(self,k)}' for k in self.__slots__])})\""
        )
        # self.class_registry[self.message_class_name].write("\n\tdef __str__(self):\n\t\treturn str(self.__dict__)\n\tdef __repr__(self):\n\t\treturn str(self)")

        self.class_registry[self.message_class_name].write(
            f"\n\tdef deserialize(self, stream, TRIM_SIZE):"
        )
        for member in member_declaration_list:
            if isinstance(member, RosMessageParser.Field_declarationContext):
                # self.class_registry[self.message_class_name].write(f"\n\t\t{member.getChild(0).getText()}")
                self.class_registry[self.message_class_name].write(
                    f"\n\t\tself.{member.getChild(1).getText()} = {self.visit(member.getChild(0))}"
                )

        # self.class_registry[self.message_class_name].write(f"\n\tdef serialize(self, outputstream):")

        # {key : getattr(self, key, None) for key in self.__slots__}")
        convert_to_dict = f"\n\tdef  to_dict(self):\n\t\t# call to_dict if it exists for the member otherwise return the member\n\t\treturn {{key : getattr(self, key, None).to_dict() if hasattr(getattr(self, key, None), 'to_dict') else getattr(self, key, None) for key in self.__slots__}}"
        self.class_registry[self.message_class_name].write(convert_to_dict)
        self.class_registry[self.message_class_name].write(
            f"\n\tdef json(self):\n\t\treturn"
            + "{"
            + ",".join(
                [
                    f"'{member[0]}': getattr(self, '{member[0]}', None)"
                    for member in declared_members
                ]
            )
            + "}"
        )

        return self.visitChildren(ctx)

    # Visit a parse tree produced by RosMessageParser#field_declaration.
    def visitField_declaration(self, ctx: RosMessageParser.Field_declarationContext):
        # self.class_registry[self.message_class_name].write("\n\t", ctx.getText(),end="")
        return self.visitChildren(ctx)

    # Visit a parse tree produced by RosMessageParser#constant_declaration.
    def visitConstant_declaration(
        self, ctx: RosMessageParser.Constant_declarationContext
    ):
        # self.class_registry[self.message_class_name].write("\nvisitConstant_declaration")
        return self.visitChildren(ctx)

    # Visit a parse tree produced by RosMessageParser#comment.
    def visitComment(self, ctx: RosMessageParser.CommentContext):
        # self.class_registry[self.message_class_name].write("\n\t", ctx.getText())
        return self.visitChildren(ctx)

    # Visit a parse tree produced by RosMessageParser#identifier.
    def visitIdentifier(self, ctx: RosMessageParser.IdentifierContext):
        # self.class_registry[self.message_class_name].write("\n ", ctx.getText(),end="\n")
        return self.visitChildren(ctx)

    # Visit a parse tree produced by RosMessageParser#type.
    def visitType(self, ctx: RosMessageParser.Type_Context):
        # assert that only child node
        assert ctx.getChildCount() == 1
        return self.visitChildren(ctx)

    # Visit a parse tree produced by RosMessageParser#ros_type.
    def visitRos_type(self, ctx: RosMessageParser.Ros_typeContext):
        class_type = ctx.getText().replace("/", ".")
        if class_type == "Header":
            # prefix with package as std_msgs
            class_type = "std_msgs.Header"
        return f"read_object(type='{class_type}', stream=stream, current_package=self.package,TRIM_SIZE=TRIM_SIZE)"

        return self.visitChildren(ctx)

    # Visit a parse tree produced by RosMessageParser#array_type.
    def visitArray_type(self, ctx: RosMessageParser.Array_typeContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by RosMessageParser#variable_array_type.
    def visitVariable_array_type(
        self, ctx: RosMessageParser.Variable_array_typeContext
    ):
        return f"read_array(type='{ctx.getChild(0).getText()}', \
stream=stream, read_element=lambda stream,TRIM_SIZE:{self.visit(ctx.getChild(0))}, length=None, package=self.package,TRIM_SIZE=TRIM_SIZE)"
        # return self.visitChildren(ctx)

    # Visit a parse tree produced by RosMessageParser#fixed_array_type.
    def visitFixed_array_type(self, ctx: RosMessageParser.Fixed_array_typeContext):
        return f"read_array(type='{ctx.getChild(0).getText()}', \
stream=stream, read_element=lambda stream:{self.visit(ctx.getChild(0))}, length={ctx.getChild(2).getText()}, package=self.package,TRIM_SIZE=TRIM_SIZE)"

        return self.visitChildren(ctx)

    # Visit a parse tree produced by RosMessageParser#integral_type.
    def visitIntegral_type(self, ctx: RosMessageParser.Integral_typeContext):
        return "read_" + ctx.getText() + "(stream,TRIM_SIZE=TRIM_SIZE)"

    # Visit a parse tree produced by RosMessageParser#floating_point_type.
    def visitFloating_point_type(
        self, ctx: RosMessageParser.Floating_point_typeContext
    ):
        return "read_" + ctx.getText() + "(stream,TRIM_SIZE=TRIM_SIZE)"

    # Visit a parse tree produced by RosMessageParser#temporal_type.
    def visitTemporal_type(self, ctx: RosMessageParser.Temporal_typeContext):
        return "read_" + ctx.getText() + "(stream,TRIM_SIZE=TRIM_SIZE)"

    # Visit a parse tree produced by RosMessageParser#string_type.
    def visitString_type(self, ctx: RosMessageParser.String_typeContext):
        return "read_" + ctx.getText() + "(stream,TRIM_SIZE=TRIM_SIZE)"

    # Visit a parse tree produced by RosMessageParser#boolean_type.
    def visitBoolean_type(self, ctx: RosMessageParser.Boolean_typeContext):
        return "read_" + ctx.getText() + "(stream,TRIM_SIZE=TRIM_SIZE)"

    # Visit a parse tree produced by RosMessageParser#sign.
    def visitSign(self, ctx: RosMessageParser.SignContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by RosMessageParser#integral_value.
    def visitIntegral_value(self, ctx: RosMessageParser.Integral_valueContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by RosMessageParser#floating_point_value.
    def visitFloating_point_value(
        self, ctx: RosMessageParser.Floating_point_valueContext
    ):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by RosMessageParser#bool_value.
    def visitBool_value(self, ctx: RosMessageParser.Bool_valueContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by RosMessageParser#string_value.
    def visitString_value(self, ctx: RosMessageParser.String_valueContext):
        return self.visitChildren(ctx)


class RosMessageParserVisitorForJsonSchema(ParseTreeVisitor):
    def __init__(self, message_class_name: str, generate_unnested_schema: bool):
        self.root_class_name = message_class_name
        self.generate_unnested_schema = generate_unnested_schema

    # Visit a parse tree produced by RosMessageParser#rosbag_input.
    def visitRosbag_input(self, ctx: RosMessageParser.Rosbag_inputContext):
        self.class_registry = {}
        # visit children in reverse order, that makes it easier since the dependencies get processed first
        for child in reversed(ctx.children[1:]):
            self.visit(child)
        self.message_class_name = self.root_class_name
        if "/" in self.message_class_name:
            self.current_package = self.message_class_name.split("/")[0]
        self.visit(ctx.children[0])
        if self.generate_unnested_schema:
            unnested_json_schema = {"$defs": {}}
            for k, v in self.class_registry.items():
                v["$id"] = k
                unnested_json_schema["$defs"][k] = v

            # remove our "root" schema from defs and put it at the top level
            unnested_json_schema.update(
                unnested_json_schema["$defs"].pop(self.root_class_name)
            )
            return unnested_json_schema
        return self.class_registry[self.root_class_name]

    # Visit a parse tree produced by RosMessageParser#rosbag_nested_message.
    def visitRosbag_nested_message(
        self, ctx: RosMessageParser.Rosbag_nested_messageContext
    ):
        self.message_class_name = ctx.getChild(1).getText()
        if "/" in self.message_class_name:
            self.current_package = self.message_class_name.split("/")[0]
        return self.visitChildren(ctx)

    def convert_ros_to_json(self, ros_type: Union[str, dict]):
        if isinstance(ros_type, str) and ros_type in self.class_registry:
            if self.generate_unnested_schema:
                return {"$ref": f"#/$defs/{ros_type}"}
            return self.class_registry[ros_type]
        else:
            return ros_type

    # Visit a parse tree produced by RosMessageParser#ros_message.
    def visitRos_message(self, ctx: RosMessageParser.Ros_messageContext):
        # collect all the non-comment children
        member_declaration_list = list(
            ctx.getChildren(
                lambda x: not isinstance(x, RosMessageParser.CommentContext)
            )
        )
        processed_members = [self.visit(member) for member in member_declaration_list]
        schema = {
            "type": "object",
            "properties": {result[0]: result[1] for result in processed_members},
        }

        if self.message_class_name not in self.class_registry:
            self.class_registry[self.message_class_name] = {}
        self.class_registry.get(self.message_class_name).update(schema)

    # Visit a parse tree produced by RosMessageParser#field_declaration.
    def visitField_declaration(self, ctx: RosMessageParser.Field_declarationContext):
        member_name = ctx.getChild(1).getText()
        member_type = self.visit(ctx.getChild(0))
        return member_name, member_type

    # Visit a parse tree produced by RosMessageParser#constant_declaration.
    def visitConstant_declaration(
        self, ctx: RosMessageParser.Constant_declarationContext
    ):
        member_name = ctx.getChild(1).getText()
        const_value = self.visit(ctx.getChild(3))
        return member_name, {
            "const": const_value,
        }

    # Visit a parse tree produced by RosMessageParser#comment.
    def visitComment(self, ctx: RosMessageParser.CommentContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by RosMessageParser#identifier.
    def visitIdentifier(self, ctx: RosMessageParser.IdentifierContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by RosMessageParser#type.
    def visitType(self, ctx: RosMessageParser.Type_Context):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by RosMessageParser#ros_type.
    def visitRos_type(self, ctx: RosMessageParser.Ros_typeContext):
        class_type = ctx.getText()
        if class_type == "Header":
            # prefix with package as std_msgs
            class_type = "std_msgs/Header"
        # return class_type
        if "/" not in class_type:
            class_type = f"{self.current_package}/{class_type}"
        return {"$ref": f"#/$defs/{class_type}"}
        # return self.class_registry[class_type]

    # Visit a parse tree produced by RosMessageParser#array_type.
    def visitArray_type(self, ctx: RosMessageParser.Array_typeContext):
        res = self.visitChildren(ctx)
        array_element_type = res[1]
        res = {
            "type": "array",
            "items": self.convert_ros_to_json(array_element_type),
        }
        if len(res) == 3:
            # fixed size array
            res["length"] = res[2]
        return res

    # Visit a parse tree produced by RosMessageParser#variable_array_type.
    def visitVariable_array_type(
        self, ctx: RosMessageParser.Variable_array_typeContext
    ):
        # array, type
        return "array", self.visit(ctx.getChild(0))

    # Visit a parse tree produced by RosMessageParser#fixed_array_type.
    def visitFixed_array_type(self, ctx: RosMessageParser.Fixed_array_typeContext):
        # array, type, and fixed length
        return "array", self.visit(ctx.getChild(0)), ctx.getChild(2).getText()

    # Visit a parse tree produced by RosMessageParser#integral_type.
    def visitIntegral_type(self, ctx: RosMessageParser.Integral_typeContext):
        return {"type": "number"}

    # Visit a parse tree produced by RosMessageParser#floating_point_type.
    def visitFloating_point_type(
        self, ctx: RosMessageParser.Floating_point_typeContext
    ):
        return {"type": "number"}

    # Visit a parse tree produced by RosMessageParser#temporal_type.
    def visitTemporal_type(self, ctx: RosMessageParser.Temporal_typeContext):
        return {
            "type": "object",
            "properties": {
                "secs": {"type": "number"},
                "nsecs": {"type": "number"},
            },
        }

    # Visit a parse tree produced by RosMessageParser#string_type.
    def visitString_type(self, ctx: RosMessageParser.String_typeContext):
        return {"type": "string"}

    # Visit a parse tree produced by RosMessageParser#boolean_type.
    def visitBoolean_type(self, ctx: RosMessageParser.Boolean_typeContext):
        return {"type": "boolean"}

    # Visit a parse tree produced by RosMessageParser#sign.
    def visitSign(self, ctx: RosMessageParser.SignContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by RosMessageParser#integral_value.
    def visitIntegral_value(self, ctx: RosMessageParser.Integral_valueContext):
        return int(ctx.getText())

    # Visit a parse tree produced by RosMessageParser#floating_point_value.
    def visitFloating_point_value(
        self, ctx: RosMessageParser.Floating_point_valueContext
    ):
        return float(ctx.getText())

    # Visit a parse tree produced by RosMessageParser#bool_value.
    def visitBool_value(self, ctx: RosMessageParser.Bool_valueContext):
        return bool(ctx.getText())

    # Visit a parse tree produced by RosMessageParser#string_value.
    def visitString_value(self, ctx: RosMessageParser.String_valueContext):
        return ctx.getText()



from enum import Enum
from typing import Optional

from pydantic import BaseModel

from lqs.interface.dsm.models.__common__ import (
    DataResponseModel,
)


class TicketType(str, Enum):
    question = "Question"
    software_issue = "Software Issue"
    feature_request = "Feature Request"
    billing = "Billing"
    documentation = "Documentation"
    other = "Other"


class Ticket(BaseModel):
    subject: str
    type: TicketType
    description: str
    email: str
    priority: int
    status: int
    product_id: Optional[int]
    custom_fields: Optional[dict]


class TicketDataResponse(DataResponseModel[Ticket]):
    pass


class TicketCreateRequest(BaseModel):
    type: TicketType
    description: str
    email: str



from lqs.interface.core import FetchInterface
from lqs.client.common import RESTInterface
import lqs.interface.core.models as models


class Fetch(FetchInterface, RESTInterface):
    service: str = "lqs"

    def __init__(self, app):
        super().__init__(app=app)

    def _api_key(self, **params):
        api_key_id = params.pop("api_key_id")
        result = self._get_resource(
            f"apiKeys/{api_key_id}", response_model=models.APIKeyDataResponse
        )
        return result

    def _callback(self, **params):
        callback_id = params.pop("callback_id")
        result = self._get_resource(
            f"callbacks/{callback_id}", response_model=models.CallbackDataResponse
        )
        return result

    def _digestion(self, **params):
        digestion_id = params.pop("digestion_id")
        result = self._get_resource(
            f"digestions/{digestion_id}", response_model=models.DigestionDataResponse
        )
        return result

    def _digestion_part(self, **kwargs):
        digestion_id = kwargs.pop("digestion_id", None)
        digestion_part_id = kwargs.pop("digestion_part_id")
        if digestion_id is None:
            resource_path = f"digestionParts/{digestion_part_id}"
        else:
            resource_path = f"digestions/{digestion_id}/parts/{digestion_part_id}"
        result = self._get_resource(
            resource_path,
            response_model=models.DigestionPartDataResponse,
        )
        return result

    def _digestion_topic(self, **params):
        digestion_id = params.pop("digestion_id", None)
        digestion_topic_id = params.pop("digestion_topic_id")
        if digestion_id is None:
            resource_path = f"digestionTopics/{digestion_topic_id}"
        else:
            resource_path = f"digestions/{digestion_id}/topics/{digestion_topic_id}"
        result = self._get_resource(
            resource_path,
            response_model=models.DigestionTopicDataResponse,
        )
        return result

    def _group(self, **params):
        group_id = params.pop("group_id")
        result = self._get_resource(
            f"groups/{group_id}", response_model=models.GroupDataResponse
        )
        return result

    def _hook(self, **params):
        workflow_id = params.pop("workflow_id", None)
        hook_id = params.pop("hook_id")
        if workflow_id is None:
            resource_path = f"hooks/{hook_id}"
        else:
            resource_path = f"workflows/{workflow_id}/hooks/{hook_id}"
        result = self._get_resource(
            resource_path,
            response_model=models.HookDataResponse,
        )
        return result

    def _ingestion(self, **params):
        ingestion_id = params.pop("ingestion_id")
        result = self._get_resource(
            f"ingestions/{ingestion_id}", response_model=models.IngestionDataResponse
        )
        return result

    def _ingestion_part(self, **params):
        ingestion_id = params.pop("ingestion_id", None)
        ingestion_part_id = params.pop("ingestion_part_id")
        if ingestion_id is None:
            resource_path = f"ingestionParts/{ingestion_part_id}"
        else:
            resource_path = f"ingestions/{ingestion_id}/parts/{ingestion_part_id}"
        result = self._get_resource(
            resource_path,
            response_model=models.IngestionPartDataResponse,
        )
        return result

    def _label(self, **params):
        label_id = params.pop("label_id")
        result = self._get_resource(
            f"labels/{label_id}", response_model=models.LabelDataResponse
        )
        return result

    def _log(self, **params):
        log_id = params.pop("log_id")
        result = self._get_resource(
            f"logs/{log_id}", response_model=models.LogDataResponse
        )
        return result

    def _log_object(self, **params):
        log_id = params.pop("log_id")
        object_key = params.pop("object_key")

        resource_path = (
            f"logs/{log_id}/objects/{object_key}"
            + self._get_url_param_string(params, [])
        )

        if params.get("redirect", False):
            offset = params.pop("offset", None)
            length = params.pop("length", None)
            headers = {}
            if offset is not None:
                if length is not None:
                    headers["Range"] = f"bytes={offset}-{offset + length - 1}"
                else:
                    if offset < 0:
                        headers["Range"] = f"bytes={offset}"
                    else:
                        headers["Range"] = f"bytes={offset}-"
            elif length is not None:
                headers["Range"] = f"bytes=0-{length - 1}"
            result = self._get_resource(
                resource_path, expected_content_type=None, additional_headers=headers
            )
        else:
            result = self._get_resource(
                resource_path, response_model=models.ObjectDataResponse
            )
        return result

    def _log_object_part(self, **params):
        log_id = params.pop("log_id")
        object_key = params.pop("object_key")
        part_number = params.pop("part_number")
        result = self._get_resource(
            f"logs/{log_id}/objects/{object_key}/parts/{part_number}",
            response_model=models.ObjectPartDataResponse,
        )
        return result

    def _me(self, **params):
        result = self._get_resource("users/me", response_model=models.MeDataResponse)
        return result

    def _object(self, **params):
        object_store_id = params.pop("object_store_id")
        object_key = params.pop("object_key")

        resource_path = (
            f"objectStores/{object_store_id}/objects/{object_key}"
            + self._get_url_param_string(params, [])
        )

        if params.get("redirect", False):
            offset = params.pop("offset", None)
            length = params.pop("length", None)
            headers = {}
            if offset is not None:
                if length is not None:
                    headers["Range"] = f"bytes={offset}-{offset + length - 1}"
                else:
                    if offset < 0:
                        headers["Range"] = f"bytes={offset}"
                    else:
                        headers["Range"] = f"bytes={offset}-"
            elif length is not None:
                headers["Range"] = f"bytes=0-{length - 1}"
            result = self._get_resource(
                resource_path, expected_content_type=None, additional_headers=headers
            )
        else:
            result = self._get_resource(
                resource_path, response_model=models.ObjectDataResponse
            )
        return result

    def _object_part(self, **params):
        raise NotImplementedError

    def _object_store(self, **params):
        object_store_id = params.pop("object_store_id")
        result = self._get_resource(
            f"objectStores/{object_store_id}",
            response_model=models.ObjectStoreDataResponse,
        )
        return result

    def _query(self, **params):
        log_id = params.pop("log_id", None)
        query_id = params.pop("query_id")
        if log_id is None:
            resource_path = f"queries/{query_id}"
        else:
            resource_path = f"logs/{log_id}/queries/{query_id}"
        result = self._get_resource(
            resource_path, response_model=models.QueryDataResponse
        )
        return result

    def _record(self, **params):
        topic_id = params.pop("topic_id")
        timestamp = params.pop("timestamp")
        url_params = self._get_url_param_string(params, [])
        result = self._get_resource(
            f"topics/{topic_id}/records/{timestamp}" + url_params,
            response_model=models.RecordDataResponse,
        )
        return result

    def _role(self, **params):
        role_id = params.pop("role_id")
        result = self._get_resource(
            f"roles/{role_id}", response_model=models.RoleDataResponse
        )
        return result

    def _tag(self, **params):
        log_id = params.pop("log_id", None)
        tag_id = params.pop("tag_id")
        if log_id is None:
            resource_path = f"tags/{tag_id}"
        else:
            resource_path = f"logs/{log_id}/tags/{tag_id}"
        result = self._get_resource(
            resource_path, response_model=models.TagDataResponse
        )
        return result

    def _topic(self, **params):
        topic_id = params.pop("topic_id")
        result = self._get_resource(
            f"topics/{topic_id}", response_model=models.TopicDataResponse
        )
        return result

    def _user(self, **params):
        user_id = params.pop("user_id")
        result = self._get_resource(
            f"users/{user_id}", response_model=models.UserDataResponse
        )
        return result

    def _workflow(self, **params):
        workflow_id = params.pop("workflow_id")
        result = self._get_resource(
            f"workflows/{workflow_id}", response_model=models.WorkflowDataResponse
        )
        return result



from typing import Optional
from uuid import UUID

from pydantic import BaseModel, Field, ConfigDict

from lqs.interface.core.models.__common__ import (
    CommonModel,
    DataResponseModel,
    PaginationModel,
    optional_field,
    Int64,
)
from lqs.interface.core import models


class Tag(CommonModel):
    label_id: UUID = Field(
        ..., description="The ID of the label associated with the tag"
    )
    log_id: UUID = Field(..., description="The ID of the log associated with the tag")
    topic_id: Optional[UUID] = Field(
        None, description="The ID of the topic associated with the tag"
    )

    note: Optional[str] = Field(
        None, description="A general note about the tag for reference."
    )
    context: Optional[dict] = Field(
        None, description="Arbitrary JSON context for the tag."
    )
    start_time: Optional[Int64] = Field(
        None, description="The timestamp of the beggining of the tagged time range."
    )
    end_time: Optional[Int64] = Field(
        None, description="The timestamp of the end of the tagged time range."
    )

    model_config = ConfigDict(
        json_schema_extra=lambda x: x["required"].extend(
            ["topic_id", "note", "context", "start_time", "end_time"]
        )
    )

    def fetch_label(self) -> "models.Label":
        return self._app.fetch.label(self.label_id).data

    def fetch_log(self) -> "models.Log":
        return self._app.fetch.log(self.log_id).data

    def fetch_topic(self) -> Optional["models.Topic"]:
        if self.topic_id is None:
            return None
        return self._app.fetch.topic(self.topic_id).data


class TagDataResponse(DataResponseModel[Tag]):
    pass


class TagListResponse(PaginationModel[Tag]):
    pass


class TagForLogCreateRequest(BaseModel):
    label_id: UUID
    topic_id: Optional[UUID] = None

    note: Optional[str] = None
    context: Optional[dict] = None
    start_time: Optional[Int64] = None
    end_time: Optional[Int64] = None


class TagCreateRequest(TagForLogCreateRequest):
    log_id: UUID


class TagUpdateRequest(BaseModel):
    label_id: UUID = optional_field
    topic_id: Optional[UUID] = optional_field

    note: Optional[str] = optional_field
    context: Optional[dict] = optional_field
    start_time: Optional[Int64] = optional_field
    end_time: Optional[Int64] = optional_field



from uuid import UUID
from typing import Optional

from lqs.common.config import CommonConfig


class RESTClientConfig(CommonConfig):
    """
    Configuration for the RESTClient class.

    Attributes:
        api_key_id (Optional[UUID]): The ID of the API key with which requests will be made.
        api_key_secret (Optional[str]): The secret of the API key with which requests will be made.
        datastore_id (Optional[UUID]): The ID of the DataStore which the client will interact with.

        api_url (Optional[str]): The base URL of the LogQS API (default: "https://api.logqs.com")
        api_endpoint_prefix (str): The prefix for the API endpoints (default: "/apps")
        dsm_api_key_id (Optional[UUID]): The ID of the API key used to interact with the DataStore Manager.
        dsm_api_key_secret (Optional[str]): The secret of the API key used to interact with the DataStore Manager.

        pretty (bool): Whether to pretty-print responses (default: False)
        verbose (bool): Whether to print verbose output (default: False)
        log_level (str): The logging level (default: "INFO")
        log_as_json (bool): Whether to log as JSON (default: False)
        dry_run (bool): Whether to run in dry-run mode (default: False)
        retry_count (int): The number of times to retry requests (default: 2)
        retry_delay (int): The delay between retries (default: 5)
        retry_aggressive (bool): Whether to use an aggressive retry strategy (default: False)
        api_request_timeout (int): The timeout for API requests (default: 60)
        additional_headers (dict[str, str]): Additional headers to include in requests (default: {})
    """

    api_key_id: Optional[UUID] = None
    api_key_secret: Optional[str] = None
    datastore_id: Optional[UUID] = None

    api_url: Optional[str] = "https://api.logqs.com"
    api_endpoint_prefix: str = "/apps"
    dsm_api_key_id: Optional[UUID] = None
    dsm_api_key_secret: Optional[str] = None

    pretty: bool = False
    verbose: bool = False
    log_level: str = "INFO"
    log_as_json: bool = False
    dry_run: bool = False
    retry_count: int = 2
    retry_delay: int = 5
    retry_aggressive: bool = False
    api_request_timeout: int = 60
    version_match_level: Optional[str] = "minor"
    raise_for_version_mismatch: bool = False
    warn_for_version_mismatch: bool = True
    additional_headers: dict[str, str] = {}
    session_id: Optional[str] = None



from typing import Optional

from pydantic import BaseModel, Field, ConfigDict

from lqs.interface.core.models.__common__ import (
    CommonModel,
    DataResponseModel,
    PaginationModel,
    optional_field,
)
from lqs.interface.core import models


class Label(CommonModel):
    value: str = Field(..., description="The value of the label")
    note: Optional[str] = Field(
        None, description="A general note about the label for reference."
    )
    category: Optional[str] = Field(
        None,
        description="The category of the label used to group labels of similar purpose.",
    )

    model_config = ConfigDict(
        json_schema_extra=lambda x: x["required"].extend(["note", "category"])
    )

    def list_tags(self, threaded=False, **kwargs) -> list["models.Tag"]:
        return self._list_all_subresources(
            list_method=self._app.list.tag,
            threaded=threaded,
            label_id=self.id,
            **kwargs,
        )


class LabelDataResponse(DataResponseModel[Label]):
    pass


class LabelListResponse(PaginationModel[Label]):
    pass


class LabelCreateRequest(BaseModel):
    value: str
    note: Optional[str] = None
    category: Optional[str] = None


class LabelUpdateRequest(BaseModel):
    value: str = optional_field
    note: Optional[str] = optional_field
    category: Optional[str] = optional_field



import json
from typing import Type, Any, Union
from logging import Logger

from lqs.common.config import CommonConfig
from lqs.common.utils import get_logger, ContextLoggerAdapter
from lqs.interface import core


class Facade:
    logger: Union[Logger, ContextLoggerAdapter]

    def setup(
        self,
        config: Union[CommonConfig, dict[str, Any], str],
        config_class: Type[CommonConfig],
    ):
        if isinstance(config, dict):
            self.config: config_class = config_class(**config)
        elif isinstance(config, str):
            self.config: config_class = config_class(**json.loads(config))
        else:
            self.config: config_class = config
        log_level = self.config.log_level or "INFO"

        self.correlation_id = getattr(self.config, "session_id", None)
        self.logger = get_logger(
            self.__class__.__name__,
            level=log_level,
            log_to_file=self.config.log_to_file,
            json_logging=self.config.log_as_json,
            correlation_id=self.correlation_id,
        )

    def json(self, req):
        """
        Convert a JSON request to a dictionary.
        """
        return self.dict(json.loads(req))

    def dict(self, req):
        """
        Parse a dictionary request to a dictionary.
        """
        action = req.get("action")
        resource = req.get("resource")
        kwargs = req.get("kwargs")

        if action is None:
            raise Exception('[BadRequest] Required field "action" is missing.')

        if resource is None:
            resource_handler = getattr(self, action)
        else:
            action_handler = getattr(self, action)
            resource_handler = getattr(action_handler, resource)

        if kwargs is None:
            return resource_handler()
        else:
            return resource_handler(**kwargs)


class CoreFacade(Facade):
    fetch: core.FetchInterface
    list: core.ListInterface
    update: core.UpdateInterface
    create: core.CreateInterface
    delete: core.DeleteInterface



from typing import Optional
from uuid import UUID

from pydantic import BaseModel, Field, ConfigDict

from lqs.interface.base.models.__common__ import (
    CommonModel,
    DataResponseModel,
    PaginationModel,
    optional_field,
)
from lqs.interface.core import models


class APIKey(CommonModel["APIKey"]):
    name: str = Field(..., description="Name of the API key for reference.")
    user_id: UUID = Field(
        ..., description="The ID of the user to which the API key belongs."
    )
    disabled: bool = Field(
        ..., description="Whether the API key is disabled and cannot be used."
    )
    secret: Optional[str] = Field(
        None,
        description="The secret key for the API key. This is only returned when the API key is created.",
    )

    model_config = ConfigDict(
        json_schema_extra=lambda x: x["required"].extend(["secret"])
    )

    def fetch_user(self) -> "models.User":
        return self._app.fetch.user(self.user_id).data


class APIKeyDataResponse(DataResponseModel[APIKey]):
    pass


class APIKeyListResponse(PaginationModel[APIKey]):
    pass


class APIKeyCreateRequest(BaseModel):
    name: str
    user_id: UUID
    disabled: bool = False


class APIKeyUpdateRequest(BaseModel):
    name: str = optional_field
    disabled: bool = optional_field



from lqs.interface.core.models.api_key import *
from lqs.interface.core.models.callback import *
from lqs.interface.core.models.digestion import *
from lqs.interface.core.models.digestion_part import *
from lqs.interface.core.models.digestion_topic import *

from lqs.interface.core.models.hook import *
from lqs.interface.core.models.ingestion import *
from lqs.interface.core.models.ingestion_part import *
from lqs.interface.core.models.label import *

from lqs.interface.core.models.record import (
    Record,
    RecordDataResponse,
    RecordListResponse,
    RecordCreateRequest,
    RecordUpdateRequest,
)
from lqs.interface.core.models.topic import (
    Topic,
    TopicDataResponse,
    TopicListResponse,
    TopicCreateRequest,
    TopicUpdateRequest,
)
from lqs.interface.core.models.log import (
    Log,
    LogDataResponse,
    LogListResponse,
    LogCreateRequest,
    LogUpdateRequest,
)
from lqs.interface.core.models.group import (
    Group,
    GroupDataResponse,
    GroupListResponse,
    GroupCreateRequest,
    GroupUpdateRequest,
)

from lqs.interface.core.models.object import *
from lqs.interface.core.models.object_store import *
from lqs.interface.core.models.query import *

from lqs.interface.core.models.role import *
from lqs.interface.core.models.tag import *

from lqs.interface.core.models.user import *
from lqs.interface.core.models.workflow import *

from lqs.interface.core.models.jsonl import *
from lqs.interface.core.models.__common__ import *


class Resources:
    APIKey = APIKey
    Callback = Callback
    Digestion = Digestion
    DigestionPart = DigestionPart
    DigestionTopic = DigestionTopic
    Group = Group
    Hook = Hook
    Ingestion = Ingestion
    IngestionPart = IngestionPart
    Label = Label
    Log = Log
    Object = Object
    ObjectStore = ObjectStore
    Query = Query
    Record = Record
    Role = Role
    Tag = Tag
    Topic = Topic
    User = User
    Workflow = Workflow
